import dataclasses
import warnings
from enum import Enum
from typing import Annotated, Any

import pydantic
import pytest

import strawberry
from strawberry.experimental.pydantic.exceptions import MissingFieldsListError
from strawberry.schema_directive import Location
from strawberry.types.base import (
    StrawberryList,
    StrawberryObjectDefinition,
    StrawberryOptional,
)
from strawberry.types.enum import StrawberryEnumDefinition
from strawberry.types.union import StrawberryUnion


def test_basic_type_all_fields():
    class User(pydantic.BaseModel):
        age: int
        password: str | None

    @strawberry.experimental.pydantic.type(User, all_fields=True)
    class UserType:
        pass

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field1, field2] = definition.fields

    assert field1.python_name == "age"
    assert field1.graphql_name is None
    assert field1.type is int

    assert field2.python_name == "password"
    assert field2.graphql_name is None
    assert isinstance(field2.type, StrawberryOptional)
    assert field2.type.of_type is str


def test_basic_type_all_fields_respects_explicit_definitions():
    class User(pydantic.BaseModel):
        age: int
        password: str | None

    # Should NOT produce a warning - explicit definitions are now respected
    with warnings.catch_warnings(record=True) as w:
        warnings.simplefilter("always")

        @strawberry.experimental.pydantic.type(User, all_fields=True)
        class UserType:
            age: strawberry.auto

        relevant_warnings = [
            warning for warning in w if "all_fields overrides" in str(warning.message)
        ]
        assert len(relevant_warnings) == 0

    # Verify the type was created correctly with both fields
    definition = UserType.__strawberry_definition__
    field_names = {f.name for f in definition.fields}
    assert field_names == {"age", "password"}


def test_basic_type_auto_fields():
    class User(pydantic.BaseModel):
        age: int
        password: str | None
        other: float

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto
        password: strawberry.auto

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field1, field2] = definition.fields

    assert field1.python_name == "age"
    assert field1.graphql_name is None
    assert field1.type is int

    assert field2.python_name == "password"
    assert field2.graphql_name is None
    assert isinstance(field2.type, StrawberryOptional)
    assert field2.type.of_type is str


def test_auto_fields_other_sentinel():
    class OtherSentinel:
        pass

    class User(pydantic.BaseModel):
        age: int
        password: str | None
        other: int

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto
        password: strawberry.auto
        other: OtherSentinel  # this should be a private field, not an auto field

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field1, field2, field3] = definition.fields

    assert field1.python_name == "age"
    assert field1.graphql_name is None
    assert field1.type is int

    assert field2.python_name == "password"
    assert field2.graphql_name is None
    assert isinstance(field2.type, StrawberryOptional)
    assert field2.type.of_type is str

    assert field3.python_name == "other"
    assert field3.graphql_name is None
    assert field3.type is OtherSentinel


def test_referencing_other_models_fails_when_not_registered():
    class Group(pydantic.BaseModel):
        name: str

    class User(pydantic.BaseModel):
        age: int
        password: str | None
        group: Group

    with pytest.raises(
        strawberry.experimental.pydantic.UnregisteredTypeException,
        match=r"Cannot find a Strawberry Type for (.*) did you forget to register it?",
    ):

        @strawberry.experimental.pydantic.type(User)
        class UserType:
            age: strawberry.auto
            password: strawberry.auto
            group: strawberry.auto


def test_referencing_other_input_models_fails_when_not_registered():
    class Group(pydantic.BaseModel):
        name: str

    class User(pydantic.BaseModel):
        age: int
        password: str | None
        group: Group

    @strawberry.experimental.pydantic.type(Group)
    class GroupType:
        name: strawberry.auto

    with pytest.raises(
        strawberry.experimental.pydantic.UnregisteredTypeException,
        match=r"Cannot find a Strawberry Type for (.*) did you forget to register it?",
    ):

        @strawberry.experimental.pydantic.input(User)
        class UserInputType:
            age: strawberry.auto
            password: strawberry.auto
            group: strawberry.auto


def test_referencing_other_registered_models():
    class Group(pydantic.BaseModel):
        name: str

    class User(pydantic.BaseModel):
        age: int
        group: Group

    @strawberry.experimental.pydantic.type(Group)
    class GroupType:
        name: strawberry.auto

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto
        group: strawberry.auto

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field1, field2] = definition.fields

    assert field1.python_name == "age"
    assert field1.type is int

    assert field2.python_name == "group"
    assert field2.type is GroupType


def test_list():
    class User(pydantic.BaseModel):
        friend_names: list[str]

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        friend_names: strawberry.auto

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field] = definition.fields

    assert field.python_name == "friend_names"
    assert isinstance(field.type, StrawberryList)
    assert field.type.of_type is str


def test_list_of_types():
    class Friend(pydantic.BaseModel):
        name: str

    class User(pydantic.BaseModel):
        friends: list[Friend | None] | None

    @strawberry.experimental.pydantic.type(Friend)
    class FriendType:
        name: strawberry.auto

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        friends: strawberry.auto

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field] = definition.fields

    assert field.python_name == "friends"
    assert isinstance(field.type, StrawberryOptional)
    assert isinstance(field.type.of_type, StrawberryList)
    assert isinstance(field.type.of_type.of_type, StrawberryOptional)
    assert field.type.of_type.of_type.of_type is FriendType


def test_basic_type_without_fields_throws_an_error():
    class User(pydantic.BaseModel):
        age: int
        password: str | None

    with pytest.raises(MissingFieldsListError):

        @strawberry.experimental.pydantic.type(User)
        class UserType:
            pass


def test_type_with_fields_coming_from_strawberry_and_pydantic():
    class User(pydantic.BaseModel):
        age: int
        password: str | None

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        name: str
        age: strawberry.auto
        password: strawberry.auto

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field1, field2, field3] = definition.fields

    assert field1.python_name == "name"
    assert field1.type is str

    assert field2.python_name == "age"
    assert field2.type is int

    assert field3.python_name == "password"
    assert isinstance(field3.type, StrawberryOptional)
    assert field3.type.of_type is str


def test_default_and_default_factory():
    class User1(pydantic.BaseModel):
        friend: str | None = "friend_value"

    @strawberry.experimental.pydantic.type(User1)
    class UserType1:
        friend: strawberry.auto

    assert UserType1().friend == "friend_value"
    assert UserType1().to_pydantic().friend == "friend_value"

    class User2(pydantic.BaseModel):
        friend: str | None = None

    @strawberry.experimental.pydantic.type(User2)
    class UserType2:
        friend: strawberry.auto

    assert UserType2().friend is None
    assert UserType2().to_pydantic().friend is None

    # Test instantiation using default_factory

    class User3(pydantic.BaseModel):
        friend: str | None = pydantic.Field(default_factory=lambda: "friend_value")

    @strawberry.experimental.pydantic.type(User3)
    class UserType3:
        friend: strawberry.auto

    assert UserType3().friend == "friend_value"
    assert UserType3().to_pydantic().friend == "friend_value"

    class User4(pydantic.BaseModel):
        friend: str | None = pydantic.Field(default_factory=lambda: None)

    @strawberry.experimental.pydantic.type(User4)
    class UserType4:
        friend: strawberry.auto

    assert UserType4().friend is None
    assert UserType4().to_pydantic().friend is None


def test_optional_and_default():
    class UserModel(pydantic.BaseModel):
        age: int
        name: str = pydantic.Field("Michael", description="The user name")
        password: str | None = pydantic.Field(default="ABC")
        passwordtwo: str | None = None
        some_list: list[str] | None = pydantic.Field(default_factory=list)
        check: bool | None = False

    @strawberry.experimental.pydantic.type(UserModel, all_fields=True)
    class User:
        pass

    definition: StrawberryObjectDefinition = User.__strawberry_definition__
    assert definition.name == "User"

    [
        age_field,
        name_field,
        password_field,
        passwordtwo_field,
        some_list_field,
        check_field,
    ] = definition.fields

    assert age_field.python_name == "age"
    assert age_field.type is int

    assert name_field.python_name == "name"
    assert name_field.type is str

    assert password_field.python_name == "password"
    assert isinstance(password_field.type, StrawberryOptional)
    assert password_field.type.of_type is str

    assert passwordtwo_field.python_name == "passwordtwo"
    assert isinstance(passwordtwo_field.type, StrawberryOptional)
    assert passwordtwo_field.type.of_type is str

    assert some_list_field.python_name == "some_list"
    assert isinstance(some_list_field.type, StrawberryOptional)
    assert isinstance(some_list_field.type.of_type, StrawberryList)
    assert some_list_field.type.of_type.of_type is str

    assert check_field.python_name == "check"
    assert isinstance(check_field.type, StrawberryOptional)
    assert check_field.type.of_type is bool


def test_type_with_fields_mutable_default():
    empty_list = []

    class User(pydantic.BaseModel):
        groups: list[str]
        friends: list[str] = empty_list

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        groups: strawberry.auto
        friends: strawberry.auto

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [groups_field, friends_field] = definition.fields

    assert groups_field.default is dataclasses.MISSING
    assert groups_field.default_factory is dataclasses.MISSING
    assert friends_field.default is dataclasses.MISSING

    # check that we really made a copy
    assert friends_field.default_factory() is not empty_list
    assert UserType(groups=["groups"]).friends is not empty_list
    UserType(groups=["groups"]).friends.append("joe")
    assert empty_list == []


@pytest.mark.xfail(
    reason=(
        "passing default values when extending types from pydantic is not"
        "supported. https://github.com/strawberry-graphql/strawberry/issues/829"
    )
)
def test_type_with_fields_coming_from_strawberry_and_pydantic_with_default():
    class User(pydantic.BaseModel):
        age: int
        password: str | None

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        name: str = "Michael"
        age: strawberry.auto
        password: strawberry.auto

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field1, field2, field3] = definition.fields

    assert field1.python_name == "age"
    assert field1.type is int

    assert field2.python_name == "password"
    assert isinstance(field2.type, StrawberryOptional)
    assert field2.type.of_type is str

    assert field3.python_name == "name"
    assert field3.type is str
    assert field3.default == "Michael"


def test_type_with_nested_fields_coming_from_strawberry_and_pydantic():
    @strawberry.type
    class Name:
        first_name: str
        last_name: str

    class User(pydantic.BaseModel):
        age: int
        password: str | None

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        name: Name
        age: strawberry.auto
        password: strawberry.auto

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field1, field2, field3] = definition.fields

    assert field1.python_name == "name"
    assert field1.type is Name

    assert field2.python_name == "age"
    assert field2.type is int

    assert field3.python_name == "password"
    assert isinstance(field3.type, StrawberryOptional)
    assert field3.type.of_type is str


def test_type_with_aliased_pydantic_field():
    class UserModel(pydantic.BaseModel):
        age_: int = pydantic.Field(..., alias="age")
        password: str | None

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        age_: strawberry.auto
        password: strawberry.auto

    definition: StrawberryObjectDefinition = User.__strawberry_definition__
    assert definition.name == "User"

    [field1, field2] = definition.fields

    assert field1.python_name == "age_"
    assert field1.type is int
    assert field1.graphql_name == "age"

    assert field2.python_name == "password"
    assert isinstance(field2.type, StrawberryOptional)
    assert field2.type.of_type is str


def test_union():
    class BranchA(pydantic.BaseModel):
        field_a: str

    class BranchB(pydantic.BaseModel):
        field_b: int

    class User(pydantic.BaseModel):
        age: int
        union_field: BranchA | BranchB

    @strawberry.experimental.pydantic.type(BranchA)
    class BranchAType:
        field_a: strawberry.auto

    @strawberry.experimental.pydantic.type(BranchB)
    class BranchBType:
        field_b: strawberry.auto

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto
        union_field: strawberry.auto

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field1, field2] = definition.fields

    assert field1.python_name == "age"
    assert field1.type is int

    assert field2.python_name == "union_field"
    assert isinstance(field2.type, StrawberryUnion)
    assert field2.type.types[0] is BranchAType
    assert field2.type.types[1] is BranchBType


def test_enum():
    @strawberry.enum
    class UserKind(Enum):
        user = 0
        admin = 1

    class User(pydantic.BaseModel):
        age: int
        kind: UserKind

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto
        kind: strawberry.auto

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field1, field2] = definition.fields

    assert field1.python_name == "age"
    assert field1.type is int

    assert field2.python_name == "kind"
    assert isinstance(field2.type, StrawberryEnumDefinition)
    assert field2.type.wrapped_cls is UserKind


def test_interface():
    class Base(pydantic.BaseModel):
        base_field: str

    class BranchA(Base):
        field_a: str

    class BranchB(Base):
        field_b: int

    class User(pydantic.BaseModel):
        age: int
        interface_field: Base

    @strawberry.experimental.pydantic.interface(Base)
    class BaseType:
        base_field: strawberry.auto

    @strawberry.experimental.pydantic.type(BranchA)
    class BranchAType(BaseType):
        field_a: strawberry.auto

    @strawberry.experimental.pydantic.type(BranchB)
    class BranchBType(BaseType):
        field_b: strawberry.auto

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto
        interface_field: strawberry.auto

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field1, field2] = definition.fields

    assert field1.python_name == "age"
    assert field1.type is int

    assert field2.python_name == "interface_field"
    assert field2.type is BaseType


def test_both_output_and_input_type():
    class Work(pydantic.BaseModel):
        time: float

    class User(pydantic.BaseModel):
        name: str
        # Note that pydantic v2 requires an explicit default of None for Optionals
        work: Work | None = None

    class Group(pydantic.BaseModel):
        users: list[User]

    # Test both definition orders
    @strawberry.experimental.pydantic.input(Work)
    class WorkInput:
        time: strawberry.auto

    @strawberry.experimental.pydantic.type(Work)
    class WorkOutput:
        time: strawberry.auto

    @strawberry.experimental.pydantic.type(User)
    class UserOutput:
        name: strawberry.auto
        work: strawberry.auto

    @strawberry.experimental.pydantic.input(User)
    class UserInput:
        name: strawberry.auto
        work: strawberry.auto

    @strawberry.experimental.pydantic.input(Group)
    class GroupInput:
        users: strawberry.auto

    @strawberry.experimental.pydantic.type(Group)
    class GroupOutput:
        users: strawberry.auto

    @strawberry.type
    class Query:
        groups: list[GroupOutput]

    @strawberry.type
    class Mutation:
        @strawberry.mutation
        def update_group(group: GroupInput) -> GroupOutput:
            pass

    # This triggers the exception from #1504
    schema = strawberry.Schema(query=Query, mutation=Mutation)
    expected_schema = """
input GroupInput {
  users: [UserInput!]!
}

type GroupOutput {
  users: [UserOutput!]!
}

type Mutation {
  updateGroup(group: GroupInput!): GroupOutput!
}

type Query {
  groups: [GroupOutput!]!
}

input UserInput {
  name: String!
  work: WorkInput = null
}

type UserOutput {
  name: String!
  work: WorkOutput
}

input WorkInput {
  time: Float!
}

type WorkOutput {
  time: Float!
}"""
    assert schema.as_str().strip() == expected_schema.strip()

    assert Group._strawberry_type == GroupOutput
    assert Group._strawberry_input_type == GroupInput
    assert User._strawberry_type == UserOutput
    assert User._strawberry_input_type == UserInput
    assert Work._strawberry_type == WorkOutput
    assert Work._strawberry_input_type == WorkInput


def test_single_field_changed_type():
    class User(pydantic.BaseModel):
        age: int

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: str

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field1] = definition.fields

    assert field1.python_name == "age"
    assert field1.graphql_name is None
    assert field1.type is str


def test_type_with_aliased_pydantic_field_changed_type():
    class UserModel(pydantic.BaseModel):
        age_: int = pydantic.Field(..., alias="age")
        password: str | None

    @strawberry.experimental.pydantic.type(UserModel)
    class User:
        age_: str
        password: strawberry.auto

    definition: StrawberryObjectDefinition = User.__strawberry_definition__
    assert definition.name == "User"

    [field1, field2] = definition.fields

    assert field1.python_name == "age_"
    assert field1.type is str
    assert field1.graphql_name == "age"

    assert field2.python_name == "password"
    assert isinstance(field2.type, StrawberryOptional)
    assert field2.type.of_type is str


def test_deprecated_fields():
    class User(pydantic.BaseModel):
        age: int
        password: str | None
        other: float

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto = strawberry.field(deprecation_reason="Because")
        password: strawberry.auto

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field1, field2] = definition.fields

    assert field1.python_name == "age"
    assert field1.graphql_name is None
    assert field1.type is int
    assert field1.deprecation_reason == "Because"

    assert field2.python_name == "password"
    assert field2.graphql_name is None
    assert isinstance(field2.type, StrawberryOptional)
    assert field2.type.of_type is str


def test_permission_classes():
    class IsAuthenticated(strawberry.BasePermission):
        message = "User is not authenticated"

        def has_permission(
            self, source: Any, info: strawberry.types.Info, **kwargs: Any
        ) -> bool:
            return False

    class User(pydantic.BaseModel):
        age: int
        password: str | None
        other: float

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto = strawberry.field(permission_classes=[IsAuthenticated])
        password: strawberry.auto

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field1, field2] = definition.fields

    assert field1.python_name == "age"
    assert field1.graphql_name is None
    assert field1.type is int
    assert field1.permission_classes == [IsAuthenticated]

    assert field2.python_name == "password"
    assert field2.graphql_name is None
    assert isinstance(field2.type, StrawberryOptional)
    assert field2.type.of_type is str


def test_field_directives():
    @strawberry.schema_directive(locations=[Location.FIELD_DEFINITION])
    class Sensitive:
        reason: str

    class User(pydantic.BaseModel):
        age: int
        password: str | None
        other: float

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto = strawberry.field(directives=[Sensitive(reason="GDPR")])
        password: strawberry.auto

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field1, field2] = definition.fields

    assert field1.python_name == "age"
    assert field1.graphql_name is None
    assert field1.type is int
    assert field1.directives == [Sensitive(reason="GDPR")]

    assert field2.python_name == "password"
    assert field2.graphql_name is None
    assert isinstance(field2.type, StrawberryOptional)
    assert field2.type.of_type is str


def test_alias_fields():
    class User(pydantic.BaseModel):
        age: int

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        age: strawberry.auto = strawberry.field(name="ageAlias")

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    field1 = definition.fields[0]

    assert field1.python_name == "age"
    assert field1.graphql_name == "ageAlias"
    assert field1.type is int


def test_alias_fields_with_use_pydantic_alias():
    class User(pydantic.BaseModel):
        age: int
        state: str = pydantic.Field(alias="statePydantic")
        country: str = pydantic.Field(alias="countryPydantic")

    @strawberry.experimental.pydantic.type(User, use_pydantic_alias=True)
    class UserType:
        age: strawberry.auto = strawberry.field(name="ageAlias")
        state: strawberry.auto = strawberry.field(name="state")
        country: strawberry.auto

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field1, field2, field3] = definition.fields

    assert field1.python_name == "age"
    assert field1.graphql_name == "ageAlias"

    assert field2.python_name == "state"
    assert field2.graphql_name == "state"

    assert field3.python_name == "country"
    assert field3.graphql_name == "countryPydantic"


def test_field_metadata():
    class User(pydantic.BaseModel):
        private: bool
        public: bool

    @strawberry.experimental.pydantic.type(User)
    class UserType:
        private: strawberry.auto = strawberry.field(metadata={"admin_only": True})
        public: strawberry.auto

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field1, field2] = definition.fields

    assert field1.python_name == "private"
    assert field1.metadata["admin_only"]

    assert field2.python_name == "public"
    assert not field2.metadata


def test_annotated():
    class User(pydantic.BaseModel):
        a: Annotated[int, "metadata"]

    @strawberry.experimental.pydantic.input(User, all_fields=True)
    class UserType:
        pass

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field] = definition.fields
    assert field.python_name == "a"
    assert field.type is int


def test_nested_annotated():
    class User(pydantic.BaseModel):
        a: Annotated[int, "metadata"] | None
        b: list[Annotated[int, "metadata"]] | None

    @strawberry.experimental.pydantic.input(User, all_fields=True)
    class UserType:
        pass

    definition: StrawberryObjectDefinition = UserType.__strawberry_definition__
    assert definition.name == "UserType"

    [field_a, field_b] = definition.fields
    assert field_a.python_name == "a"
    assert isinstance(field_a.type, StrawberryOptional)
    assert field_a.type.of_type is int

    assert field_b.python_name == "b"
    assert isinstance(field_b.type, StrawberryOptional)
    assert isinstance(field_b.type.of_type, StrawberryList)
    assert field_b.type.of_type.of_type is int



import contextlib
import os
import re
from collections import defaultdict
from collections.abc import Generator
from dataclasses import dataclass
from pathlib import Path

import pytest
import rich
import rich.console
from _pytest.nodes import Item
from pluggy._result import _Result
from rich.traceback import Traceback

from strawberry.exceptions import StrawberryException, UnableToFindExceptionSource

WORKSPACE_FOLDER = Path(__file__).parents[2]
DOCS_FOLDER = WORKSPACE_FOLDER / "docs/errors"


@dataclass
class Result:
    text: str
    raised_exception: StrawberryException


@contextlib.contextmanager
def suppress_output(verbosity_level: int = 0) -> Generator[None, None, None]:
    if verbosity_level >= 2:
        yield

        return

    with (
        Path(os.devnull).open("w", encoding="utf-8") as devnull,
        contextlib.redirect_stdout(devnull),
    ):
        yield


class StrawberryExceptionsPlugin:
    def __init__(self, verbosity_level: int) -> None:
        self._info: defaultdict[type[StrawberryException], list[Result]] = defaultdict(
            list
        )
        self.verbosity_level = verbosity_level

    @pytest.hookimpl(hookwrapper=True)
    def pytest_runtest_call(self, item: Item) -> Generator[None, _Result, None]:
        __tracebackhide__ = True

        outcome = yield

        self._check_strawberry_exception(item, outcome)

    def _check_strawberry_exception(self, item: Item, outcome: _Result) -> None:
        __tracebackhide__ = True

        raises_marker = item.get_closest_marker("raises_strawberry_exception")

        if raises_marker is None:
            return

        exception = raises_marker.args[0]
        match = raises_marker.kwargs.get("match", None)

        if not issubclass(exception, StrawberryException):
            pytest.fail(f"{exception} is not a StrawberryException")

        raised_exception = outcome.excinfo[1] if outcome.excinfo else None

        # This plugin needs to work around the other hooks, see:
        # https://docs.pytest.org/en/7.1.x/how-to/writing_hook_functions.html#hookwrapper-executing-around-other-hooks
        outcome.force_result(None)

        if raised_exception is None:
            failure_message = f"Expected exception {exception}, but it did not raise"

            pytest.fail(failure_message, pytrace=False)

        if not isinstance(raised_exception, exception):
            failure_message = (
                f"Expected exception {exception}, but raised {raised_exception}"
            )

            raise raised_exception

        raised_message = str(raised_exception)

        self._collect_exception(item.name, raised_exception)

        if match is not None and not re.match(match, raised_message):
            failure_message = (
                f'"{match}" does not match raised message "{raised_message}"'
            )

            if self.verbosity_level >= 1:
                print(f"Exception: {exception}")  # noqa: T201

            pytest.fail(failure_message, pytrace=False)

    def _collect_exception(
        self, test_name: str, raised_exception: StrawberryException
    ) -> None:
        console = rich.console.Console(record=True, width=120)

        with suppress_output(self.verbosity_level):
            try:
                console.print(raised_exception)
            except UnableToFindExceptionSource:
                traceback = Traceback(
                    Traceback.extract(
                        raised_exception.__class__,
                        raised_exception,
                        raised_exception.__traceback__,
                    ),
                    max_frames=10,
                )
                console.print(traceback)

        exception_text = console.export_text()

        text = f"## {test_name}\n"

        if exception_text.strip() == "None":
            text += "No exception raised\n"
        else:
            text += f"\n``````\n{exception_text.strip()}\n``````\n\n"

        documentation_path = DOCS_FOLDER / f"{raised_exception.documentation_path}.md"

        if not documentation_path.exists():
            pytest.fail(
                f"{documentation_path.relative_to(WORKSPACE_FOLDER)} does not exist",
                pytrace=False,
            )

        self._info[raised_exception.__class__].append(
            Result(text=text, raised_exception=raised_exception)
        )

    def pytest_sessionfinish(self):
        summary_path = os.environ.get("GITHUB_STEP_SUMMARY", None)

        if not summary_path:
            return

        markdown = ""

        for exception_class, info in self._info.items():
            title = " ".join(re.findall("[a-zA-Z][^A-Z]*", exception_class.__name__))

            markdown += f"# {title}\n\n"
            markdown += (
                f"Documentation URL: {info[0].raised_exception.documentation_url}\n\n"
            )

            markdown += "\n".join([result.text for result in info])

        with Path(summary_path).open("w") as f:
            f.write(markdown)


def pytest_configure(config):
    config.pluginmanager.register(
        StrawberryExceptionsPlugin(verbosity_level=config.getoption("verbose")),
        "strawberry_exceptions",
    )

    config.addinivalue_line(
        "markers",
        "raises_strawberry_exception: expect to raise a strawberry exception.",
    )



from typing import Literal

import pytest

from tests.views.schema import schema

from .clients.base import HttpClient


@pytest.mark.parametrize("header_value", ["text/html", "*/*"])
@pytest.mark.parametrize("graphql_ide", ["graphiql", "apollo-sandbox", "pathfinder"])
async def test_renders_graphql_ide(
    header_value: str,
    http_client_class: type[HttpClient],
    graphql_ide: Literal["graphiql", "apollo-sandbox", "pathfinder"],
):
    http_client = http_client_class(schema, graphql_ide=graphql_ide)

    response = await http_client.get("/graphql", headers={"Accept": header_value})
    content_type = response.headers.get(
        "content-type", response.headers.get("Content-Type", "")
    )

    assert response.status_code == 200
    assert "text/html" in content_type
    assert "<title>Strawberry" in response.text

    if graphql_ide == "apollo-sandbox":
        assert "embeddable-sandbox.cdn.apollographql" in response.text

    if graphql_ide == "pathfinder":
        assert "@pathfinder-ide/react" in response.text

    if graphql_ide == "graphiql":
        assert "unpkg.com/graphiql" in response.text


async def test_does_not_render_graphiql_if_wrong_accept(
    http_client_class: type[HttpClient],
):
    http_client = http_client_class(schema)
    response = await http_client.get("/graphql", headers={"Accept": "text/xml"})

    # THIS might need to be changed to 404

    assert response.status_code == 400


@pytest.mark.parametrize("graphql_ide", [False, None])
async def test_renders_graphiql_disabled(
    http_client_class: type[HttpClient],
    graphql_ide: bool | None,
):
    http_client = http_client_class(schema, graphql_ide=graphql_ide)
    response = await http_client.get("/graphql", headers={"Accept": "text/html"})

    assert response.status_code == 404



from typing import List, Optional

class OperationNameResult:
    with_inputs: bool

class PersonInput:
    name: str
    age: Optional[int] = None

class ExampleInput:
    id: str
    name: str
    age: int
    person: Optional[PersonInput]
    people: list[PersonInput]
    optional_people: Optional[list[PersonInput]]

class OperationNameVariables:
    id: Optional[str]
    input: ExampleInput
    ids: list[str]
    ids2: Optional[list[Optional[str]]]
    ids3: Optional[list[Optional[list[Optional[str]]]]]



from __future__ import annotations

import json
import os
import subprocess
import tempfile
from typing import TypedDict, cast

from .result import Result, ResultType


class PyrightCLIResult(TypedDict):
    version: str
    time: str
    generalDiagnostics: list[GeneralDiagnostic]
    summary: Summary


class GeneralDiagnostic(TypedDict):
    file: str
    severity: str
    message: str
    range: Range


class Range(TypedDict):
    start: EndOrStart
    end: EndOrStart


class EndOrStart(TypedDict):
    line: int
    character: int


class Summary(TypedDict):
    filesAnalyzed: int
    errorCount: int
    warningCount: int
    informationCount: int
    timeInSec: float


def run_pyright(code: str, strict: bool = True) -> list[Result]:
    if strict:
        code = "# pyright: strict\n" + code

    with tempfile.NamedTemporaryFile("w", suffix=".py", delete=False) as f:
        f.write(code)

    process_result = subprocess.run(
        ["pyright", "--outputjson", f.name], capture_output=True, check=False
    )
    assert not process_result.stderr.decode("utf-8")

    os.unlink(f.name)  # noqa: PTH108

    pyright_result: PyrightCLIResult = json.loads(process_result.stdout.decode("utf-8"))

    result = [
        Result(
            type=cast("ResultType", diagnostic["severity"].strip()),
            message=diagnostic["message"].strip(),
            line=diagnostic["range"]["start"]["line"],
            column=diagnostic["range"]["start"]["character"] + 1,
        )
        for diagnostic in pyright_result["generalDiagnostics"]
    ]

    # make sure that results are sorted by line and column and then message
    result.sort(key=lambda x: (x.line, x.column, x.message))

    return result



from strawberry.utils.aio import (
    aenumerate,
    aislice,
    asyncgen_to_list,
    resolve_awaitable,
)


async def test_aenumerate():
    async def gen():
        yield "a"
        yield "b"
        yield "c"
        yield "d"

    res = [(i, v) async for i, v in aenumerate(gen())]
    assert res == [(0, "a"), (1, "b"), (2, "c"), (3, "d")]


async def test_aslice():
    async def gen():
        yield "a"
        yield "b"
        raise AssertionError("should never be called")  # pragma: no cover
        yield "c"  # pragma: no cover

    res = []
    async for v in aislice(gen(), 0, 2):
        res.append(v)  # noqa: PERF401

    assert res == ["a", "b"]


async def test_aislice_empty_generator():
    async def gen():
        if False:  # pragma: no cover
            yield "should not be returned"
            raise AssertionError("should never be called")

    res = []
    async for v in aislice(gen(), 0, 2):
        res.append(v)  # noqa: PERF401

    assert res == []


async def test_aislice_empty_slice():
    async def gen():
        if False:  # pragma: no cover
            yield "should not be returned"
            raise AssertionError("should never be called")

    res = []
    async for v in aislice(gen(), 0, 0):
        res.append(v)  # noqa: PERF401

    assert res == []


async def test_aislice_with_step():
    async def gen():
        yield "a"
        yield "b"
        yield "c"
        raise AssertionError("should never be called")  # pragma: no cover
        yield "d"  # pragma: no cover
        yield "e"  # pragma: no cover

    res = []
    async for v in aislice(gen(), 0, 4, 2):
        res.append(v)  # noqa: PERF401

    assert res == ["a", "c"]


async def test_asyncgen_to_list():
    async def gen():
        yield "a"
        yield "b"
        yield "c"

    assert await asyncgen_to_list(gen()) == ["a", "b", "c"]


async def test_resolve_awaitable():
    async def awaitable():
        return 1

    assert await resolve_awaitable(awaitable(), lambda v: v + 1) == 2



import sys
from typing import Annotated, Generic, TypeVar

import strawberry

T = TypeVar("T")


@strawberry.type
class TypeC:
    name: str


@strawberry.type
class Edge(Generic[T]):
    @strawberry.field
    def node(self) -> T:  # type: ignore
        ...


@strawberry.type
class Query:
    type_a: Edge[TypeC]
    type_b: Edge[Annotated["TypeC", strawberry.lazy("tests.schema.test_lazy.type_c")]]


if __name__ == "__main__":
    schema = strawberry.Schema(query=Query)
    sys.stdout.write(f"{schema.as_str()}\n")



import textwrap
from enum import Enum
from typing import Any, NoReturn

import pytest

import strawberry
from strawberry import Info
from strawberry.directive import DirectiveLocation, DirectiveValue
from strawberry.extensions import SchemaExtension
from strawberry.schema.config import StrawberryConfig
from strawberry.types.base import get_object_definition
from strawberry.utils.await_maybe import await_maybe


def test_supports_default_directives():
    @strawberry.type
    class Person:
        name: str = "Jess"
        points: int = 2000

    @strawberry.type
    class Query:
        @strawberry.field
        def person(self) -> Person:
            return Person()

    query = """query ($includePoints: Boolean!){
        person {
            name
            points @include(if: $includePoints)
        }
    }"""

    schema = strawberry.Schema(query=Query)
    result = schema.execute_sync(
        query,
        variable_values={"includePoints": False},
        context_value={"username": "foo"},
    )

    assert not result.errors
    assert result.data
    assert result.data["person"] == {"name": "Jess"}

    query = """query ($skipPoints: Boolean!){
        person {
            name
            points @skip(if: $skipPoints)
        }
    }"""

    schema = strawberry.Schema(query=Query)
    result = schema.execute_sync(query, variable_values={"skipPoints": False})

    assert not result.errors
    assert result.data
    assert result.data["person"] == {"name": "Jess", "points": 2000}


@pytest.mark.asyncio
async def test_supports_default_directives_async():
    @strawberry.type
    class Person:
        name: str = "Jess"
        points: int = 2000

    @strawberry.type
    class Query:
        @strawberry.field
        def person(self) -> Person:
            return Person()

    query = """query ($includePoints: Boolean!){
        person {
            name
            points @include(if: $includePoints)
        }
    }"""

    schema = strawberry.Schema(query=Query)
    result = await schema.execute(query, variable_values={"includePoints": False})

    assert not result.errors
    assert result.data
    assert result.data["person"] == {"name": "Jess"}

    query = """query ($skipPoints: Boolean!){
        person {
            name
            points @skip(if: $skipPoints)
        }
    }"""

    schema = strawberry.Schema(query=Query)
    result = await schema.execute(query, variable_values={"skipPoints": False})

    assert not result.errors
    assert result.data
    assert result.data["person"] == {"name": "Jess", "points": 2000}


def test_can_declare_directives():
    @strawberry.type
    class Query:
        @strawberry.field
        def cake(self) -> str:
            return "made_in_switzerland"

    @strawberry.directive(
        locations=[DirectiveLocation.FIELD], description="Make string uppercase"
    )
    def uppercase(value: DirectiveValue[str], example: str):
        return value.upper()

    schema = strawberry.Schema(query=Query, directives=[uppercase])

    expected_schema = '''
    """Make string uppercase"""
    directive @uppercase(example: String!) on FIELD

    type Query {
      cake: String!
    }
    '''

    assert schema.as_str() == textwrap.dedent(expected_schema).strip()

    result = schema.execute_sync('query { cake @uppercase(example: "foo") }')
    assert result.errors is None
    assert result.data == {"cake": "MADE_IN_SWITZERLAND"}


def test_directive_arguments_without_value_param():
    """Regression test for Strawberry Issue #1666.

    https://github.com/strawberry-graphql/strawberry/issues/1666
    """

    @strawberry.type
    class Query:
        cake: str = "victoria sponge"

    @strawberry.directive(
        locations=[DirectiveLocation.FIELD],
        description="Don't actually like cake? try ice cream instead",
    )
    def ice_cream(flavor: str):
        return f"{flavor} ice cream"

    schema = strawberry.Schema(query=Query, directives=[ice_cream])

    expected_schema = '''
    """Don't actually like cake? try ice cream instead"""
    directive @iceCream(flavor: String!) on FIELD

    type Query {
      cake: String!
    }
    '''

    assert schema.as_str() == textwrap.dedent(expected_schema).strip()

    query = 'query { cake @iceCream(flavor: "strawberry") }'
    result = schema.execute_sync(query, root_value=Query())

    assert result.data == {"cake": "strawberry ice cream"}


def test_runs_directives():
    @strawberry.type
    class Person:
        name: str = "Jess"

    @strawberry.type
    class Query:
        @strawberry.field
        def person(self) -> Person:
            return Person()

    @strawberry.directive(
        locations=[DirectiveLocation.FIELD], description="Make string uppercase"
    )
    def turn_uppercase(value: DirectiveValue[str]):
        return value.upper()

    @strawberry.directive(locations=[DirectiveLocation.FIELD])
    def replace(value: DirectiveValue[str], old: str, new: str):
        return value.replace(old, new)

    schema = strawberry.Schema(query=Query, directives=[turn_uppercase, replace])

    query = """query People($identified: Boolean!){
        person {
            name @turnUppercase
        }
        jess: person {
            name @replace(old: "Jess", new: "Jessica")
        }
        johnDoe: person {
            name @replace(old: "Jess", new: "John") @include(if: $identified)
        }
    }"""

    result = schema.execute_sync(query, variable_values={"identified": False})

    assert not result.errors
    assert result.data
    assert result.data["person"]["name"] == "JESS"
    assert result.data["jess"]["name"] == "Jessica"
    assert result.data["johnDoe"].get("name") is None


def test_runs_directives_camel_case_off():
    @strawberry.type
    class Person:
        name: str = "Jess"

    @strawberry.type
    class Query:
        @strawberry.field
        def person(self) -> Person:
            return Person()

    @strawberry.directive(
        locations=[DirectiveLocation.FIELD], description="Make string uppercase"
    )
    def turn_uppercase(value: DirectiveValue[str]):
        return value.upper()

    @strawberry.directive(locations=[DirectiveLocation.FIELD])
    def replace(value: DirectiveValue[str], old: str, new: str):
        return value.replace(old, new)

    schema = strawberry.Schema(
        query=Query,
        directives=[turn_uppercase, replace],
        config=StrawberryConfig(auto_camel_case=False),
    )

    query = """query People($identified: Boolean!){
        person {
            name @turn_uppercase
        }
        jess: person {
            name @replace(old: "Jess", new: "Jessica")
        }
        johnDoe: person {
            name @replace(old: "Jess", new: "John") @include(if: $identified)
        }
    }"""

    result = schema.execute_sync(query, variable_values={"identified": False})

    assert not result.errors
    assert result.data
    assert result.data["person"]["name"] == "JESS"
    assert result.data["jess"]["name"] == "Jessica"
    assert result.data["johnDoe"].get("name") is None


@pytest.mark.asyncio
async def test_runs_directives_async():
    @strawberry.type
    class Person:
        name: str = "Jess"

    @strawberry.type
    class Query:
        @strawberry.field
        def person(self) -> Person:
            return Person()

    @strawberry.directive(
        locations=[DirectiveLocation.FIELD], description="Make string uppercase"
    )
    async def uppercase(value: DirectiveValue[str]):
        return value.upper()

    schema = strawberry.Schema(query=Query, directives=[uppercase])

    query = """{
        person {
            name @uppercase
        }
    }"""

    result = await schema.execute(query, variable_values={"identified": False})

    assert not result.errors
    assert result.data
    assert result.data["person"]["name"] == "JESS"


@pytest.mark.xfail
def test_runs_directives_with_list_params():
    @strawberry.type
    class Person:
        name: str = "Jess"

    @strawberry.type
    class Query:
        @strawberry.field
        def person(self) -> Person:
            return Person()

    @strawberry.directive(locations=[DirectiveLocation.FIELD])
    def replace(value: DirectiveValue[str], old_list: list[str], new: str):
        for old in old_list:
            value = value.replace(old, new)

        return value

    schema = strawberry.Schema(query=Query, directives=[replace])

    query = """query People {
        person {
            name @replace(oldList: ["J", "e", "s", "s"], new: "John")
        }
    }"""

    result = schema.execute_sync(query, variable_values={"identified": False})

    assert not result.errors
    assert result.data
    assert result.data["person"]["name"] == "JESS"


def test_runs_directives_with_extensions():
    @strawberry.type
    class Person:
        name: str = "Jess"

    @strawberry.type
    class Query:
        @strawberry.field
        def person(self) -> Person:
            return Person()

    @strawberry.directive(
        locations=[DirectiveLocation.FIELD], description="Make string uppercase"
    )
    def uppercase(value: DirectiveValue[str]):
        return value.upper()

    class ExampleExtension(SchemaExtension):
        def resolve(self, _next, root, info, *args: str, **kwargs: Any):
            return _next(root, info, *args, **kwargs)

    schema = strawberry.Schema(
        query=Query, directives=[uppercase], extensions=[ExampleExtension]
    )

    query = """query {
        person {
            name @uppercase
        }
    }"""

    result = schema.execute_sync(query)

    assert not result.errors
    assert result.data
    assert result.data["person"]["name"] == "JESS"


@pytest.mark.asyncio
async def test_runs_directives_with_extensions_async():
    @strawberry.type
    class Person:
        name: str = "Jess"

    @strawberry.type
    class Query:
        @strawberry.field
        async def person(self) -> Person:
            return Person()

    @strawberry.directive(
        locations=[DirectiveLocation.FIELD], description="Make string uppercase"
    )
    def uppercase(value: DirectiveValue[str]):
        return value.upper()

    class ExampleExtension(SchemaExtension):
        async def resolve(self, _next, root, info, *args: str, **kwargs: Any):
            return await await_maybe(_next(root, info, *args, **kwargs))

    schema = strawberry.Schema(
        query=Query, directives=[uppercase], extensions=[ExampleExtension]
    )

    query = """query {
        person {
            name @uppercase
        }
    }"""

    result = await schema.execute(query)

    assert not result.errors
    assert result.data
    assert result.data["person"]["name"] == "JESS"


@pytest.fixture
def info_directive_schema() -> strawberry.Schema:
    """Returns a schema with directive that validates if info is recieved."""

    @strawberry.enum
    class Locale(Enum):
        EN: str = "EN"
        NL: str = "NL"

    greetings: dict[Locale, str] = {
        Locale.EN: "Hello {username}",
        Locale.NL: "Hallo {username}",
    }

    @strawberry.type
    class Query:
        @strawberry.field
        def greeting_template(self, locale: Locale = Locale.EN) -> str:
            return greetings[locale]

    field = get_object_definition(Query, strict=True).fields[0]

    @strawberry.directive(
        locations=[DirectiveLocation.FIELD],
        description="Interpolate string on the server from context data",
    )
    def interpolate(value: DirectiveValue[str], info: strawberry.Info):
        try:
            assert isinstance(info, strawberry.Info)
            assert info._field is field
            return value.format(**info.context["userdata"])
        except KeyError:
            return value

    return strawberry.Schema(query=Query, directives=[interpolate])


def test_info_directive_schema(info_directive_schema: strawberry.Schema):
    expected_schema = '''
    """Interpolate string on the server from context data"""
    directive @interpolate on FIELD

    enum Locale {
      EN
      NL
    }

    type Query {
      greetingTemplate(locale: Locale! = EN): String!
    }
    '''

    assert textwrap.dedent(expected_schema).strip() == str(info_directive_schema)


def test_info_directive(info_directive_schema: strawberry.Schema):
    query = "query { greetingTemplate @interpolate }"
    result = info_directive_schema.execute_sync(
        query, context_value={"userdata": {"username": "Foo"}}
    )
    assert result.data == {"greetingTemplate": "Hello Foo"}


@pytest.mark.asyncio
async def test_info_directive_async(info_directive_schema: strawberry.Schema):
    query = "query { greetingTemplate @interpolate }"
    result = await info_directive_schema.execute(
        query, context_value={"userdata": {"username": "Foo"}}
    )
    assert result.data == {"greetingTemplate": "Hello Foo"}


def test_directive_value():
    """Tests if directive value is detected by type instead of by arg-name `value`."""

    @strawberry.type
    class Cake:
        frosting: str | None = None
        flavor: str = "Chocolate"

    @strawberry.type
    class Query:
        @strawberry.field
        def cake(self) -> Cake:
            return Cake()

    @strawberry.directive(
        locations=[DirectiveLocation.FIELD],
        description="Add frostring with ``flavor`` to a cake.",
    )
    def add_frosting(flavor: str, v: DirectiveValue[Cake], value: str):
        assert isinstance(v, Cake)
        assert value == "foo"  # Check if value can be used as an argument
        v.frosting = flavor
        return v

    schema = strawberry.Schema(query=Query, directives=[add_frosting])
    result = schema.execute_sync(
        """query {
            cake @addFrosting(flavor: "Vanilla", value: "foo") {
                frosting
                flavor
            }
        }
        """
    )
    assert result.data == {"cake": {"frosting": "Vanilla", "flavor": "Chocolate"}}


# Defined in module scope to allow the FowardRef to be resolvable with eval
@strawberry.directive(
    locations=[DirectiveLocation.FIELD],
    description="Add frostring with ``flavor`` to a cake.",
)
def add_frosting(flavor: str, v: DirectiveValue["Cake"], value: str) -> "Cake":
    assert isinstance(v, Cake)
    assert value == "foo"
    v.frosting = flavor
    return v


@strawberry.type
class Query:
    @strawberry.field
    def cake(self) -> "Cake":
        return Cake()


@strawberry.type
class Cake:
    frosting: str | None = None
    flavor: str = "Chocolate"


def test_directive_value_forward_ref():
    """Tests if directive value by type works with PEP-563."""
    schema = strawberry.Schema(query=Query, directives=[add_frosting])
    result = schema.execute_sync(
        """query {
            cake @addFrosting(flavor: "Vanilla", value: "foo") {
                frosting
                flavor
            }
        }
        """
    )
    assert result.data == {"cake": {"frosting": "Vanilla", "flavor": "Chocolate"}}


def test_name_first_directive_value():
    @strawberry.type
    class Query:
        @strawberry.field
        def greeting(self) -> str:
            return "Hi"

    @strawberry.directive(locations=[DirectiveLocation.FIELD])
    def personalize_greeting(value: str, v: DirectiveValue[str]):
        assert v == "Hi"
        return f"{v} {value}"

    schema = strawberry.Schema(Query, directives=[personalize_greeting])
    result = schema.execute_sync('{ greeting @personalizeGreeting(value: "Bar")}')

    assert result.data is not None
    assert not result.errors
    assert result.data["greeting"] == "Hi Bar"


@pytest.mark.asyncio
async def test_directive_list_argument() -> NoReturn:
    @strawberry.type
    class Query:
        @strawberry.field
        def greeting(self) -> str:
            return "Hi"

    @strawberry.directive(locations=[DirectiveLocation.FIELD])
    def append_names(value: DirectiveValue[str], names: list[str]):
        assert isinstance(names, list)
        return f"{value} {', '.join(names)}"

    schema = strawberry.Schema(query=Query, directives=[append_names])

    result = await schema.execute(
        'query { greeting @appendNames(names: ["foo", "bar"])}'
    )

    assert result.errors is None
    assert result.data
    assert result.data["greeting"] == "Hi foo, bar"


def test_directives_with_custom_types():
    @strawberry.input
    class DirectiveInput:
        example: str

    @strawberry.type
    class Query:
        @strawberry.field
        def cake(self) -> str:
            return "made_in_switzerland"

    @strawberry.directive(
        locations=[DirectiveLocation.FIELD], description="Make string uppercase"
    )
    def uppercase(value: DirectiveValue[str], input: DirectiveInput):
        return value.upper()

    schema = strawberry.Schema(query=Query, directives=[uppercase])

    expected_schema = '''
    """Make string uppercase"""
    directive @uppercase(input: DirectiveInput!) on FIELD

    input DirectiveInput {
      example: String!
    }

    type Query {
      cake: String!
    }
    '''

    assert schema.as_str() == textwrap.dedent(expected_schema).strip()

    result = schema.execute_sync('query { cake @uppercase(input: { example: "foo" }) }')
    assert result.errors is None
    assert result.data == {"cake": "MADE_IN_SWITZERLAND"}


def test_directives_with_scalar():
    DirectiveInput = strawberry.scalar(str, name="DirectiveInput")

    @strawberry.type
    class Query:
        @strawberry.field
        def cake(self) -> str:
            return "made_in_switzerland"

    @strawberry.directive(
        locations=[DirectiveLocation.FIELD], description="Make string uppercase"
    )
    def uppercase(value: DirectiveValue[str], input: DirectiveInput):
        return value.upper()

    schema = strawberry.Schema(query=Query, directives=[uppercase])

    expected_schema = '''
    """Make string uppercase"""
    directive @uppercase(input: DirectiveInput!) on FIELD

    scalar DirectiveInput

    type Query {
      cake: String!
    }
    '''

    assert schema.as_str() == textwrap.dedent(expected_schema).strip()

    result = schema.execute_sync('query { cake @uppercase(input: "foo") }')
    assert result.errors is None
    assert result.data == {"cake": "MADE_IN_SWITZERLAND"}


@pytest.mark.asyncio
async def test_directive_with_custom_info_class() -> NoReturn:
    @strawberry.type
    class Query:
        @strawberry.field
        def greeting(self) -> str:
            return "Hi"

    class CustomInfo(Info):
        test: str = "foo"

    @strawberry.directive(locations=[DirectiveLocation.FIELD])
    def append_names(value: DirectiveValue[str], names: list[str], info: CustomInfo):
        assert isinstance(names, list)
        assert isinstance(info, CustomInfo)
        assert Info in type(info).__bases__  # Explicitly check it's not Info.
        assert info.test == "foo"
        return f"{value} {', '.join(names)}"

    schema = strawberry.Schema(
        query=Query,
        directives=[append_names],
        config=StrawberryConfig(info_class=CustomInfo),
    )

    result = await schema.execute(
        'query { greeting @appendNames(names: ["foo", "bar"])}'
    )

    assert result.errors is None
    assert result.data
    assert result.data["greeting"] == "Hi foo, bar"



import asyncio  # noqa: INP001
import random

import strawberry
from strawberry.schema.config import StrawberryConfig


@strawberry.type
class Author:
    id: strawberry.ID
    name: str


@strawberry.type
class Comment:
    id: strawberry.ID
    content: str

    @strawberry.field
    async def author(self) -> Author:
        await asyncio.sleep(random.uniform(0, 0.5))  # noqa: S311
        return Author(id=strawberry.ID("Author:1"), name="John Doe")


@strawberry.type
class BlogPost:
    id: strawberry.ID
    title: str
    content: str

    @strawberry.field
    async def comments(self) -> strawberry.Streamable[Comment]:
        for x in range(5):
            await asyncio.sleep(random.uniform(0, 0.5))  # noqa: S311

            yield Comment(id=strawberry.ID(f"Comment:{x}"), content="Great post!")


@strawberry.type
class Query:
    @strawberry.field
    async def hello(self, delay: float = 0) -> str:
        await asyncio.sleep(delay)
        return "Hello, world!"

    @strawberry.field
    async def blog_post(self, id: strawberry.ID) -> BlogPost:
        return BlogPost(id=id, title="My Blog Post", content="This is my blog post.")


schema = strawberry.Schema(
    query=Query,
    config=StrawberryConfig(
        enable_experimental_incremental_execution=True,
    ),
)



import typing
from collections.abc import AsyncGenerator
from typing import Any

import pytest

import strawberry

if typing.TYPE_CHECKING:
    from strawberry.extensions.tracing.datadog import DatadogTracingExtension


@pytest.fixture
def ddtrace_version_2(mocker):
    ddtrace_mock = mocker.MagicMock()
    ddtrace_mock.__version__ = "2.20.0"
    mocker.patch.dict("sys.modules", ddtrace=ddtrace_mock)
    return ddtrace_mock


@pytest.fixture
def ddtrace_version_3(mocker):
    ddtrace_mock = mocker.MagicMock()
    ddtrace_mock.__version__ = "3.0.0"
    mocker.patch.dict("sys.modules", ddtrace=ddtrace_mock)

    trace_mock = mocker.MagicMock()
    mocker.patch.dict("sys.modules", {"ddtrace.trace": trace_mock})
    return trace_mock


@pytest.fixture(params=["ddtrace_version_2", "ddtrace_version_3"])
def datadog_extension(request) -> tuple[type["DatadogTracingExtension"], Any]:
    fixture_name = request.param
    ddtrace_mock = request.getfixturevalue(fixture_name)

    from strawberry.extensions.tracing.datadog import DatadogTracingExtension

    return DatadogTracingExtension, ddtrace_mock


@pytest.fixture(params=["ddtrace_version_2", "ddtrace_version_3"])
def datadog_extension_sync(request) -> tuple[type["DatadogTracingExtension"], Any]:
    fixture_name = request.param
    ddtrace_mock = request.getfixturevalue(fixture_name)

    from strawberry.extensions.tracing.datadog import DatadogTracingExtensionSync

    return DatadogTracingExtensionSync, ddtrace_mock


@strawberry.type
class Person:
    name: str = "Jack"


@strawberry.type
class Query:
    @strawberry.field
    def person(self) -> Person:
        return Person()

    @strawberry.field
    async def person_async(self) -> Person:
        return Person()


@strawberry.type
class Mutation:
    @strawberry.mutation
    def say_hi(self) -> str:
        return "hello"


@strawberry.type
class Subscription:
    @strawberry.field
    async def on_hi(self) -> AsyncGenerator[str, None]:
        yield "Hello"


# TODO: this test could be improved by passing a custom tracer to the datadog extension
# and maybe we could unify datadog and opentelemetry extensions by doing that


@pytest.mark.asyncio
async def test_datadog_tracer(datadog_extension, mocker):
    extension, mock = datadog_extension

    schema = strawberry.Schema(
        query=Query,
        mutation=Mutation,
        extensions=[extension],
    )

    query = """
        query {
            personAsync {
                name
            }
        }
    """

    await schema.execute(query)

    mock.tracer.assert_has_calls(
        [
            mocker.call.trace(
                "Anonymous Query",
                resource="63a280256ca4e8514e06cf90b30c8c3a",
                span_type="graphql",
                service="strawberry",
            ),
            mocker.call.trace().set_tag("graphql.operation_name", None),
            mocker.call.trace().set_tag("graphql.operation_type", "query"),
            mocker.call.trace("Parsing", span_type="graphql"),
            mocker.call.trace().finish(),
            mocker.call.trace("Validation", span_type="graphql"),
            mocker.call.trace().finish(),
            mocker.call.trace("Resolving: Query.personAsync", span_type="graphql"),
            mocker.call.trace().__enter__(),
            mocker.call.trace()
            .__enter__()
            .set_tag("graphql.field_name", "personAsync"),
            mocker.call.trace().__enter__().set_tag("graphql.parent_type", "Query"),
            mocker.call.trace()
            .__enter__()
            .set_tag("graphql.field_path", "Query.personAsync"),
            mocker.call.trace().__enter__().set_tag("graphql.path", "personAsync"),
            mocker.call.trace().__exit__(None, None, None),
            mocker.call.trace().finish(),
        ]
    )


@pytest.mark.asyncio
async def test_uses_operation_name_and_hash(datadog_extension):
    extension, mock = datadog_extension

    schema = strawberry.Schema(query=Query, mutation=Mutation, extensions=[extension])

    query = """
        query MyExampleQuery {
            person {
                name
            }
        }
    """

    await schema.execute(query, operation_name="MyExampleQuery")

    mock.tracer.trace.assert_any_call(
        "MyExampleQuery",
        resource="MyExampleQuery:efe8d7247ee8136f45e3824c2768b155",
        span_type="graphql",
        service="strawberry",
    )


@pytest.mark.asyncio
async def test_uses_operation_type(datadog_extension):
    extension, mock = datadog_extension

    schema = strawberry.Schema(query=Query, mutation=Mutation, extensions=[extension])

    query = """
        mutation MyMutation {
            sayHi
        }
    """

    await schema.execute(query, operation_name="MyMutation")
    mock.tracer.trace().set_tag.assert_any_call("graphql.operation_type", "mutation")


@pytest.mark.asyncio
async def test_uses_operation_subscription(datadog_extension):
    extension, mock = datadog_extension

    schema = strawberry.Schema(query=Query, mutation=Mutation, extensions=[extension])

    query = """
        subscription MySubscription {
            onHi
        }
    """

    await schema.execute(query, operation_name="MySubscription")
    mock.tracer.trace().set_tag.assert_any_call(
        "graphql.operation_type", "subscription"
    )


def test_datadog_tracer_sync(datadog_extension_sync, mocker):
    extension, mock = datadog_extension_sync
    schema = strawberry.Schema(query=Query, mutation=Mutation, extensions=[extension])

    query = """
        query {
            person {
                name
            }
        }
    """

    schema.execute_sync(query)

    mock.tracer.assert_has_calls(
        [
            mocker.call.trace(
                "Anonymous Query",
                resource="659edba9e6ac9c20d03da1b2d0f9a956",
                span_type="graphql",
                service="strawberry",
            ),
            mocker.call.trace().set_tag("graphql.operation_name", None),
            mocker.call.trace().set_tag("graphql.operation_type", "query"),
            mocker.call.trace("Parsing", span_type="graphql"),
            mocker.call.trace().finish(),
            mocker.call.trace("Validation", span_type="graphql"),
            mocker.call.trace().finish(),
            mocker.call.trace("Resolving: Query.person", span_type="graphql"),
            mocker.call.trace().__enter__(),
            mocker.call.trace().__enter__().set_tag("graphql.field_name", "person"),
            mocker.call.trace().__enter__().set_tag("graphql.parent_type", "Query"),
            mocker.call.trace()
            .__enter__()
            .set_tag("graphql.field_path", "Query.person"),
            mocker.call.trace().__enter__().set_tag("graphql.path", "person"),
            mocker.call.trace().__exit__(None, None, None),
            mocker.call.trace().finish(),
        ]
    )


def test_uses_operation_name_and_hash_sync(datadog_extension_sync):
    extension, mock = datadog_extension_sync

    schema = strawberry.Schema(query=Query, mutation=Mutation, extensions=[extension])

    query = """
        query MyExampleQuery {
            person {
                name
            }
        }
    """

    schema.execute_sync(query, operation_name="MyExampleQuery")

    mock.tracer.trace.assert_any_call(
        "MyExampleQuery",
        resource="MyExampleQuery:efe8d7247ee8136f45e3824c2768b155",
        span_type="graphql",
        service="strawberry",
    )


def test_uses_operation_type_sync(datadog_extension_sync):
    extension, mock = datadog_extension_sync

    schema = strawberry.Schema(query=Query, mutation=Mutation, extensions=[extension])

    query = """
        mutation MyMutation {
            sayHi
        }
    """

    schema.execute_sync(query, operation_name="MyMutation")

    mock.tracer.trace().set_tag.assert_any_call("graphql.operation_type", "mutation")


@pytest.mark.asyncio
async def test_create_span_override(datadog_extension):
    from strawberry.extensions.tracing.datadog import LifecycleStep

    extension, mock = datadog_extension

    class CustomExtension(extension):
        def create_span(
            self,
            lifecycle_step: LifecycleStep,
            name: str,
            **kwargs,  # noqa: ANN003
        ):
            span = super().create_span(lifecycle_step, name, **kwargs)
            if lifecycle_step == LifecycleStep.OPERATION:
                span.set_tag("graphql.query", self.execution_context.query)
            return span

    schema = strawberry.Schema(
        query=Query,
        mutation=Mutation,
        extensions=[CustomExtension],
    )

    query = """
        query {
            personAsync {
                name
            }
        }
    """

    await schema.execute(query)

    mock.tracer.trace().set_tag.assert_any_call("graphql.query", query)


@pytest.mark.asyncio
async def test_uses_query_missing_operation_if_no_query(datadog_extension, mocker):
    """Avoid regression of https://github.com/strawberry-graphql/strawberry/issues/3150"""
    extension, mock = datadog_extension

    schema = strawberry.Schema(query=Query, mutation=Mutation, extensions=[extension])

    # A missing query error is expected here, but the extension will run anyways
    with pytest.raises(strawberry.exceptions.MissingQueryError):
        await schema.execute(None)

    mock.tracer.assert_has_calls(
        [
            mocker.call.trace(
                "Anonymous Query",
                resource="query_missing",
                span_type="graphql",
                service="strawberry",
            ),
            mocker.call.trace().set_tag("graphql.operation_name", None),
            mocker.call.trace().set_tag("graphql.operation_type", "query_missing"),
        ]
    )



import asyncio
from asyncio.futures import Future
from collections.abc import Awaitable, Callable
from typing import Any, Optional, cast

import pytest
from pytest_mock import MockerFixture

from strawberry.dataloader import AbstractCache, DataLoader
from strawberry.exceptions import WrongNumberOfResultsReturned

IDXType = Callable[[list[int]], Awaitable[list[int]]]


async def idx(keys: list[int]) -> list[int]:
    return keys


@pytest.mark.asyncio
async def test_loading():
    loader = DataLoader(load_fn=idx)

    value_a = await loader.load(1)
    value_b = await loader.load(2)
    value_c = await loader.load(3)

    assert value_a == 1
    assert value_b == 2
    assert value_c == 3

    values = await loader.load_many([1, 2, 3, 4, 5, 6])

    assert values == [1, 2, 3, 4, 5, 6]


@pytest.mark.asyncio
async def test_gathering(mocker: MockerFixture):
    mock_loader = mocker.Mock(side_effect=idx)

    loader = DataLoader(load_fn=cast("IDXType", mock_loader))

    [value_a, value_b, value_c] = await asyncio.gather(
        loader.load(1),
        loader.load(2),
        loader.load(3),
    )

    mock_loader.assert_called_once_with([1, 2, 3])

    assert value_a == 1
    assert value_b == 2
    assert value_c == 3


@pytest.mark.asyncio
async def test_max_batch_size(mocker: MockerFixture):
    mock_loader = mocker.Mock(side_effect=idx)

    loader = DataLoader(load_fn=cast("IDXType", mock_loader), max_batch_size=2)

    [value_a, value_b, value_c] = await asyncio.gather(
        loader.load(1),
        loader.load(2),
        loader.load(3),
    )

    mock_loader.assert_has_calls([mocker.call([1, 2]), mocker.call([3])])  # type: ignore

    assert value_a == 1
    assert value_b == 2
    assert value_c == 3


@pytest.mark.asyncio
async def test_error():
    async def idx(keys: list[int]) -> list[int | ValueError]:
        return [ValueError()]

    loader = DataLoader(load_fn=idx)

    with pytest.raises(ValueError):
        await loader.load(1)


@pytest.mark.asyncio
async def test_error_and_values():
    async def idx(keys: list[int]) -> list[int | ValueError]:
        return [2] if keys == [2] else [ValueError()]

    loader = DataLoader(load_fn=idx)

    with pytest.raises(ValueError):
        await loader.load(1)

    assert await loader.load(2) == 2


@pytest.mark.asyncio
async def test_when_raising_error_in_loader():
    async def idx(keys: list[int]) -> list[int | ValueError]:
        raise ValueError

    loader = DataLoader(load_fn=idx)

    with pytest.raises(ValueError):
        await loader.load(1)

    with pytest.raises(ValueError):
        await asyncio.gather(
            loader.load(1),
            loader.load(2),
            loader.load(3),
        )


@pytest.mark.asyncio
async def test_returning_wrong_number_of_results():
    async def idx(keys: list[int]) -> list[int]:
        return [1, 2]

    loader = DataLoader(load_fn=idx)

    with pytest.raises(
        WrongNumberOfResultsReturned,
        match=(
            "Received wrong number of results in dataloader, expected: 1, received: 2"
        ),
    ):
        await loader.load(1)


@pytest.mark.asyncio
async def test_caches_by_id(mocker: MockerFixture):
    mock_loader = mocker.Mock(side_effect=idx)

    loader = DataLoader(load_fn=cast("IDXType", mock_loader), cache=True)

    a = loader.load(1)
    b = loader.load(1)

    assert a == b

    assert await a == 1
    assert await b == 1

    mock_loader.assert_called_once_with([1])


@pytest.mark.asyncio
async def test_caches_by_id_when_loading_many(mocker: MockerFixture):
    mock_loader = mocker.Mock(side_effect=idx)

    loader = DataLoader(load_fn=cast("IDXType", mock_loader), cache=True)

    a = loader.load(1)
    b = loader.load(1)

    assert a == b

    assert await asyncio.gather(a, b) == [1, 1]

    mock_loader.assert_called_once_with([1])


@pytest.mark.asyncio
async def test_cache_disabled(mocker: MockerFixture):
    mock_loader = mocker.Mock(side_effect=idx)

    loader = DataLoader(load_fn=cast("IDXType", mock_loader), cache=False)

    a = loader.load(1)
    b = loader.load(1)

    assert a != b

    assert await a == 1
    assert await b == 1

    mock_loader.assert_has_calls([mocker.call([1, 1])])  # type: ignore


@pytest.mark.asyncio
async def test_cache_disabled_immediate_await(mocker: MockerFixture):
    mock_loader = mocker.Mock(side_effect=idx)

    loader = DataLoader(load_fn=cast("IDXType", mock_loader), cache=False)

    a = await loader.load(1)
    b = await loader.load(1)

    assert a == b

    mock_loader.assert_has_calls([mocker.call([1]), mocker.call([1])])  # type: ignore


@pytest.mark.asyncio
async def test_prime():
    async def idx(keys: list[int | float]) -> list[int | float]:
        assert keys, "At least one key must be specified"
        return keys

    loader = DataLoader(load_fn=idx)

    # Basic behavior intact
    a1 = loader.load(1)
    assert await a1 == 1

    # Prime doesn't overrides value
    loader.prime(1, 1.1)
    loader.prime(2, 2.1)
    b1 = loader.load(1)
    b2 = loader.load(2)
    assert await b1 == 1
    assert await b2 == 2.1

    # Unless you tell it to
    loader.prime(1, 1.2, force=True)
    loader.prime(2, 2.2, force=True)
    b1 = loader.load(1)
    b2 = loader.load(2)
    assert await b1 == 1.2
    assert await b2 == 2.2

    # Preset will override pending values, but not cached values
    c2 = loader.load(2)  # This is in cache
    c3 = loader.load(3)  # This is pending
    loader.prime_many({2: 2.3, 3: 3.3}, force=True)
    assert await c2 == 2.2
    assert await c3 == 3.3

    # If we prime all keys in a batch, the load_fn is never called
    # (See assertion in idx)
    c4 = loader.load(4)
    loader.prime_many({4: 4.4})
    assert await c4 == 4.4

    # Yield to ensure the last batch has been dispatched,
    # despite all values being primed
    await asyncio.sleep(0)


@pytest.mark.asyncio
async def test_prime_nocache():
    async def idx(keys: list[int | float]) -> list[int | float]:
        assert keys, "At least one key must be specified"
        return keys

    loader = DataLoader(load_fn=idx, cache=False)

    # Primed value is ignored
    loader.prime(1, 1.1)
    a1 = loader.load(1)
    assert await a1 == 1

    # Unless it affects pending value in the current batch
    b1 = loader.load(2)
    loader.prime(2, 2.2)
    assert await b1 == 2.2

    # Yield to ensure the last batch has been dispatched,
    # despite all values being primed
    await asyncio.sleep(0)


@pytest.mark.asyncio
async def test_clear():
    batch_num = 0

    async def idx(keys: list[int]) -> list[tuple[int, int]]:
        """Maps key => (key, batch_num)"""
        nonlocal batch_num
        batch_num += 1
        return [(key, batch_num) for key in keys]

    loader = DataLoader(load_fn=idx)

    assert await loader.load_many([1, 2, 3]) == [(1, 1), (2, 1), (3, 1)]

    loader.clear(1)

    assert await loader.load_many([1, 2, 3]) == [(1, 2), (2, 1), (3, 1)]

    loader.clear_many([1, 2])

    assert await loader.load_many([1, 2, 3]) == [(1, 3), (2, 3), (3, 1)]

    loader.clear_all()

    assert await loader.load_many([1, 2, 3]) == [(1, 4), (2, 4), (3, 4)]


@pytest.mark.asyncio
async def test_clear_nocache():
    batch_num = 0

    async def idx(keys: list[int]) -> list[tuple[int, int]]:
        """Maps key => (key, batch_num)"""
        nonlocal batch_num
        batch_num += 1
        return [(key, batch_num) for key in keys]

    loader = DataLoader(load_fn=idx, cache=False)

    assert await loader.load_many([1, 2, 3]) == [(1, 1), (2, 1), (3, 1)]

    loader.clear(1)

    assert await loader.load_many([1, 2, 3]) == [(1, 2), (2, 2), (3, 2)]

    loader.clear_many([1, 2])

    assert await loader.load_many([1, 2, 3]) == [(1, 3), (2, 3), (3, 3)]

    loader.clear_all()

    assert await loader.load_many([1, 2, 3]) == [(1, 4), (2, 4), (3, 4)]


@pytest.mark.asyncio
async def test_dont_dispatch_cancelled():
    async def idx(keys: list[int]) -> list[int]:
        await asyncio.sleep(0.2)
        return keys

    loader = DataLoader(load_fn=idx)

    value_a = await loader.load(1)
    # value_b will be cancelled by hand
    value_b = cast("Future[Any]", loader.load(2))
    value_b.cancel()
    # value_c will be cancelled by the timeout
    value_c = cast("Future[Any]", loader.load(3))
    with pytest.raises(asyncio.TimeoutError):
        await asyncio.wait_for(value_c, 0.1)
    value_d = await loader.load(4)

    assert value_a == 1
    assert value_d == 4

    # 2 can still be used here because a new future will be created for it
    values = await loader.load_many([1, 2, 3, 4, 5, 6])
    assert values == [1, 2, 3, 4, 5, 6]

    with pytest.raises(asyncio.CancelledError):
        value_b.result()
    with pytest.raises(asyncio.CancelledError):
        value_c.result()  # pyright: ignore

    # Try single loading results again to make sure the cancelled
    # futures are not being reused
    value_a = await loader.load(1)
    value_b = await loader.load(2)
    value_c = await loader.load(3)
    value_d = await loader.load(4)

    assert value_a == 1
    assert value_b == 2
    assert value_c == 3
    assert value_d == 4


@pytest.mark.asyncio
async def test_cancelled_future_with_failing_loader():
    """When the load function raises and some futures are already cancelled,
    the error handler should skip cancelled futures instead of raising
    InvalidStateError.

    This reproduces a real-world scenario where a client disconnects
    mid-request: some DataLoader futures get cancelled by the event loop,
    then the batch load function fails (e.g. because the executor is torn
    down). The error handler must not crash trying to set_exception on
    the already-cancelled futures.
    """

    async def failing_loader(keys: list[int]) -> list[int]:
        raise RuntimeError("executor is broken")

    loader = DataLoader(load_fn=failing_loader, cache=False)

    future_a = cast("Future[Any]", loader.load(1))
    future_b = cast("Future[Any]", loader.load(2))
    future_c = cast("Future[Any]", loader.load(3))

    # Simulate a client disconnect cancelling one of the futures before
    # the batch dispatches
    future_a.cancel()

    # Let the event loop dispatch the batch — this must not raise
    # InvalidStateError from trying to set_exception on the cancelled future.
    # Two yields: one for call_soon to fire create_task, one for the task to run.
    await asyncio.sleep(0)
    await asyncio.sleep(0)

    # The cancelled future should remain cleanly cancelled — not broken by
    # an InvalidStateError inside dispatch_batch
    assert future_a.cancelled()
    with pytest.raises(asyncio.CancelledError):
        future_a.result()

    # The non-cancelled futures should have received the loader's exception
    # (without the fix, dispatch_batch crashes on future_a before reaching these)
    with pytest.raises(RuntimeError, match="executor is broken"):
        future_b.result()
    with pytest.raises(RuntimeError, match="executor is broken"):
        future_c.result()


@pytest.mark.asyncio
async def test_cache_override():
    class TestCache(AbstractCache[int, int]):
        def __init__(self):
            self.cache: dict[int, Future[int]] = {}

        def get(self, key: int) -> Optional["Future[int]"]:
            return self.cache.get(key)

        def set(self, key: int, value: "Future[int]") -> None:
            self.cache[key] = value

        def delete(self, key: int) -> None:
            del self.cache[key]

        def clear(self) -> None:
            self.cache.clear()

    custom_cache = TestCache()
    loader = DataLoader(load_fn=idx, cache_map=custom_cache)

    await loader.load(1)
    await loader.load(2)
    await loader.load(3)

    assert len(custom_cache.cache) == 3
    assert await custom_cache.cache[1] == 1
    assert await custom_cache.cache[2] == 2
    assert await custom_cache.cache[3] == 3

    loader.clear(1)
    assert len(custom_cache.cache) == 2
    assert sorted(custom_cache.cache.keys()) == [2, 3]

    loader.clear_all()
    assert len(custom_cache.cache) == 0
    assert not list(custom_cache.cache.keys())

    await loader.load(1)
    await loader.load(2)
    await loader.load(3)

    loader.clear_many([1, 2])
    assert len(custom_cache.cache) == 1
    assert list(custom_cache.cache.keys()) == [3]

    data = await loader.load(3)
    assert data == 3

    loader.prime(3, 4)
    assert await custom_cache.cache[3] == 3

    loader.prime(3, 4, True)
    assert await custom_cache.cache[3] == 4

    with pytest.raises(TypeError, match="unhashable type: 'list'"):
        await loader.load([1, 2, 3])  # type: ignore

    data = await loader.load((1, 2, 3))  # type: ignore
    assert await custom_cache.get((1, 2, 3)) == data  # type: ignore


@pytest.mark.asyncio
async def test_custom_cache_key_fn():
    def custom_cache_key(key: list[int]) -> str:
        return ",".join(str(k) for k in key)

    loader = DataLoader(load_fn=idx, cache_key_fn=custom_cache_key)
    data = await loader.load([1, 2, "test"])
    assert data == [1, 2, "test"]


@pytest.mark.asyncio
async def test_user_class_custom_cache_key_fn():
    class CustomData:
        def __init__(self, custom_id: int, name: str):
            self.id: int = custom_id
            self.name: str = name

    def custom_cache_key(key: CustomData) -> int:
        return key.id

    loader = DataLoader(load_fn=idx, cache_key_fn=custom_cache_key)
    data1 = await loader.load(CustomData(1, "Nick"))
    data2 = await loader.load(CustomData(1, "Nick"))
    assert data1 == data2

    data2 = await loader.load(CustomData(2, "Jane"))
    assert data1 != data2


def test_works_when_created_in_a_different_loop(mocker: MockerFixture):
    mock_loader = mocker.Mock(side_effect=idx)
    loader = DataLoader(load_fn=cast("IDXType", mock_loader), cache=False)

    loop = asyncio.new_event_loop()

    async def run():
        return await loader.load(1)

    data = loop.run_until_complete(run())

    assert data == 1

    mock_loader.assert_called_once_with([1])



import textwrap

from strawberry.schema_codegen import codegen


def test_union():
    schema = """
    union User = Admin | Client

    type Admin {
        name: String!
    }

    type Client {
        name: String!
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry
        from typing import Annotated

        @strawberry.type
        class Admin:
            name: str

        @strawberry.type
        class Client:
            name: str

        User = Annotated[Admin | Client, strawberry.union(name="User")]
        """
    ).strip()

    assert codegen(schema).strip() == expected



import asyncio
import datetime
import random
from datetime import date
from typing import cast

import pytest
from pytest_codspeed.plugin import BenchmarkFixture

import strawberry
from strawberry.scalars import ID


@pytest.mark.benchmark
def test_execute(benchmark: BenchmarkFixture):
    birthday = datetime.datetime.now()
    pets = ("cat", "shark", "dog", "lama")

    @strawberry.type
    class Pet:
        id: int
        name: str

    @strawberry.type
    class Patron:
        id: int
        name: str
        age: int
        birthday: date
        tags: list[str]

        @strawberry.field
        def pets(self) -> list[Pet]:
            return [
                Pet(
                    id=i,
                    name=random.choice(pets),  # noqa: S311
                )
                for i in range(5)
            ]

    @strawberry.type
    class Query:
        @strawberry.field
        def patrons(self) -> list[Patron]:
            return [
                Patron(
                    id=i,
                    name="Patrick",
                    age=100,
                    birthday=birthday,
                    tags=["go", "ajax"],
                )
                for i in range(1000)
            ]

    schema = strawberry.Schema(query=Query)

    query = """
        query something{
          patrons {
            id
            name
            age
            birthday
            tags
            pets {
                id
                name
            }
          }
        }
    """

    def run():
        return asyncio.run(schema.execute(query))

    benchmark(run)


@pytest.mark.parametrize("ntypes", [2**k for k in range(0, 13, 4)])
def test_interface_performance(benchmark: BenchmarkFixture, ntypes: int):
    @strawberry.interface
    class Item:
        id: ID

    CONCRETE_TYPES: list[type[Item]] = [
        strawberry.type(type(f"Item{i}", (Item,), {})) for i in range(ntypes)
    ]

    @strawberry.type
    class Query:
        items: list[Item]

    schema = strawberry.Schema(query=Query, types=CONCRETE_TYPES)
    query = "query { items { id } }"

    def run():
        return asyncio.run(
            schema.execute(
                query,
                root_value=Query(
                    items=[
                        CONCRETE_TYPES[i % ntypes](id=cast("ID", i))
                        for i in range(1000)
                    ]
                ),
            )
        )

    benchmark(run)



from unittest import mock

import pytest

from tests.views.schema import schema


def _fake_asgi():
    return lambda: None


@mock.patch("strawberry.channels.router.GraphQLHTTPConsumer.as_asgi")
@mock.patch("strawberry.channels.router.GraphQLWSConsumer.as_asgi")
@pytest.mark.parametrize("pattern", ["^graphql", "^foo"])
def test_included_paths(ws_asgi: mock.Mock, http_asgi: mock.Mock, pattern: str):
    from strawberry.channels.router import GraphQLProtocolTypeRouter

    http_ret = _fake_asgi()
    http_asgi.return_value = http_ret

    ws_ret = _fake_asgi()
    ws_asgi.return_value = ws_ret

    router = GraphQLProtocolTypeRouter(schema, url_pattern=pattern)
    assert set(router.application_mapping) == {"http", "websocket"}

    assert len(router.application_mapping["http"].routes) == 1
    http_route = router.application_mapping["http"].routes[0]
    assert http_route.pattern._regex == pattern
    assert http_route.callback is http_ret

    assert len(router.application_mapping["websocket"].routes) == 1
    http_route = router.application_mapping["websocket"].routes[0]
    assert http_route.pattern._regex == pattern
    assert http_route.callback is ws_ret


@mock.patch("strawberry.channels.router.GraphQLHTTPConsumer.as_asgi")
@mock.patch("strawberry.channels.router.GraphQLWSConsumer.as_asgi")
@pytest.mark.parametrize("pattern", ["^graphql", "^foo"])
def test_included_paths_with_django_app(
    ws_asgi: mock.Mock,
    http_asgi: mock.Mock,
    pattern: str,
):
    from strawberry.channels.router import GraphQLProtocolTypeRouter

    http_ret = _fake_asgi()
    http_asgi.return_value = http_ret

    ws_ret = _fake_asgi()
    ws_asgi.return_value = ws_ret

    django_app = _fake_asgi()
    router = GraphQLProtocolTypeRouter(
        schema,
        django_application=django_app,
        url_pattern=pattern,
    )
    assert set(router.application_mapping) == {"http", "websocket"}

    assert len(router.application_mapping["http"].routes) == 2
    http_route = router.application_mapping["http"].routes[0]
    assert http_route.pattern._regex == pattern
    assert http_route.callback is http_ret

    django_route = router.application_mapping["http"].routes[1]
    assert django_route.pattern._regex == "^"
    assert django_route.callback is django_app

    assert len(router.application_mapping["websocket"].routes) == 1
    http_route = router.application_mapping["websocket"].routes[0]
    assert http_route.pattern._regex == pattern
    assert http_route.callback is ws_ret



from typing import Annotated, Generic, TypeVar, Union

import pytest

import strawberry
from strawberry.annotation import StrawberryAnnotation
from strawberry.exceptions import InvalidUnionTypeError
from strawberry.types.base import get_object_definition
from strawberry.types.union import StrawberryUnion, union


def test_python_union():
    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Error:
        name: str

    annotation = StrawberryAnnotation(Union[User, Error])
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryUnion)
    assert resolved.types == (User, Error)

    assert resolved == StrawberryUnion(
        name=None,
        type_annotations=(StrawberryAnnotation(User), StrawberryAnnotation(Error)),
    )
    assert resolved == Union[User, Error]


def test_python_union_short_syntax():
    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Error:
        name: str

    annotation = StrawberryAnnotation(User | Error)
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryUnion)
    assert resolved.types == (User, Error)

    assert resolved == StrawberryUnion(
        name=None,
        type_annotations=(StrawberryAnnotation(User), StrawberryAnnotation(Error)),
    )
    assert resolved == Union[User, Error]


def test_strawberry_union():
    @strawberry.type
    class User:
        name: str

    @strawberry.type
    class Error:
        name: str

    cool_union = Annotated[User | Error, union(name="CoolUnion")]
    annotation = StrawberryAnnotation(cool_union)
    resolved = annotation.resolve()

    assert isinstance(resolved, StrawberryUnion)
    assert resolved.types == (User, Error)

    assert resolved == StrawberryUnion(
        name="CoolUnion",
        type_annotations=(StrawberryAnnotation(User), StrawberryAnnotation(Error)),
    )


def test_union_with_generic():
    T = TypeVar("T")

    @strawberry.type
    class Error:
        message: str

    @strawberry.type
    class Edge(Generic[T]):
        node: T

    Result = Annotated[Error | Edge[str], strawberry.union("Result")]

    strawberry_union = StrawberryAnnotation(Result).resolve()

    assert isinstance(strawberry_union, StrawberryUnion)
    assert strawberry_union.graphql_name == "Result"
    assert strawberry_union.types[0] == Error

    assert (
        get_object_definition(strawberry_union.types[1], strict=True).is_graphql_generic
        is False
    )


@pytest.mark.raises_strawberry_exception(
    InvalidUnionTypeError, match="Type `int` cannot be used in a GraphQL Union"
)
def test_error_with_scalar_types():
    Something = Annotated[
        int | str | float | bool,
        strawberry.union("Something"),
    ]

    @strawberry.type
    class Query:
        something: Something

    strawberry.Schema(query=Query)


@pytest.mark.raises_strawberry_exception(
    InvalidUnionTypeError, match="Type `int` cannot be used in a GraphQL Union"
)
def test_error_with_scalar_types_pipe():
    # TODO: using Something as the name of the union makes the source finder
    # use the union type defined above
    Something2 = Annotated[
        int | str | float | bool,
        strawberry.union("Something2"),
    ]

    @strawberry.type
    class Query:
        something: Something2

    strawberry.Schema(query=Query)



import textwrap
from typing import Annotated, Generic, NewType, TypeVar

import pytest

import strawberry
from strawberry.schema.config import StrawberryConfig
from strawberry.types.base import StrawberryList, StrawberryOptional
from strawberry.types.enum import StrawberryEnumDefinition
from strawberry.types.union import StrawberryUnion

T = TypeVar("T")
K = TypeVar("K")
V = TypeVar("V")


Enum = StrawberryEnumDefinition(None, name="Enum", values=[], description=None)  # type: ignore
CustomInt = strawberry.scalar(NewType("CustomInt", int))


@strawberry.type
class TypeA:
    name: str


@strawberry.type
class TypeB:
    age: int


@pytest.mark.parametrize(
    ("types", "expected_name"),
    [
        ([StrawberryList(str)], "StrListExample"),
        ([StrawberryList(StrawberryList(str))], "StrListListExample"),
        ([StrawberryOptional(StrawberryList(str))], "StrListOptionalExample"),
        ([StrawberryList(StrawberryOptional(str))], "StrOptionalListExample"),
        ([StrawberryList(Enum)], "EnumListExample"),
        ([StrawberryUnion("Union", (TypeA, TypeB))], "UnionExample"),  # pyright: ignore
        ([TypeA], "TypeAExample"),
        ([CustomInt], "CustomIntExample"),
        ([TypeA, TypeB], "TypeATypeBExample"),
        (
            [
                TypeA,
                Annotated[
                    "TypeB", strawberry.lazy("tests.objects.generics.test_names")
                ],
            ],
            "TypeATypeBExample",
        ),
    ],
)
def test_name_generation(types, expected_name):
    config = StrawberryConfig()

    @strawberry.type
    class Example(Generic[T]):
        a: T

    type_definition = Example.__strawberry_definition__  # type: ignore

    assert config.name_converter.from_generic(type_definition, types) == expected_name


def test_nested_generics():
    config = StrawberryConfig()

    @strawberry.type
    class Edge(Generic[T]):
        node: T

    @strawberry.type
    class Connection(Generic[T]):
        edges: list[T]

    type_definition = Connection.__strawberry_definition__  # type: ignore

    assert (
        config.name_converter.from_generic(
            type_definition,
            [
                Edge[int],
            ],
        )
        == "IntEdgeConnection"
    )


def test_nested_generics_aliases_with_schema():
    """This tests is similar to the previous test, but it also tests against
    the schema, since the resolution of the type name might be different.
    """
    config = StrawberryConfig()

    @strawberry.type
    class Value(Generic[T]):
        value: T

    @strawberry.type
    class DictItem(Generic[K, V]):
        key: K
        value: V

    type_definition = Value.__strawberry_definition__  # type: ignore

    assert (
        config.name_converter.from_generic(
            type_definition,
            [
                StrawberryList(DictItem[int, str]),
            ],
        )
        == "IntStrDictItemListValue"
    )

    @strawberry.type
    class Query:
        d: Value[list[DictItem[int, str]]]

    schema = strawberry.Schema(query=Query)

    expected = textwrap.dedent(
        """
        type IntStrDictItem {
          key: Int!
          value: String!
        }

        type IntStrDictItemListValue {
          value: [IntStrDictItem!]!
        }

        type Query {
          d: IntStrDictItemListValue!
        }
        """
    ).strip()

    assert str(schema) == expected



import textwrap

from strawberry.schema_codegen import codegen


def test_extend_query():
    schema = """
    extend type Query {
        world: String!
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.type
        class Query:
            world: str

        schema = strawberry.Schema(query=Query)
        """
    ).strip()

    assert codegen(schema).strip() == expected


def test_extend_mutation():
    schema = """
    extend type Mutation {
        world: String!
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.type
        class Mutation:
            world: str

        schema = strawberry.Schema(mutation=Mutation)
        """
    ).strip()

    assert codegen(schema).strip() == expected


def test_extend_subscription():
    schema = """
    extend type Subscription {
        world: String!
    }
    """

    expected = textwrap.dedent(
        """
        from __future__ import annotations
        import strawberry

        @strawberry.type
        class Subscription:
            world: str

        schema = strawberry.Schema(subscription=Subscription)
        """
    ).strip()

    assert codegen(schema).strip() == expected



import sys
from typing import Annotated, Generic, TypeVar

import strawberry

T = TypeVar("T")


class Mixin(Generic[T]):
    node: T


@strawberry.type
class TypeD(Mixin[int]):
    name: str


@strawberry.type
class Query:
    type_d_1: TypeD
    type_d: Annotated["TypeD", strawberry.lazy("tests.schema.test_lazy.type_d")]


if __name__ == "__main__":
    schema = strawberry.Schema(query=Query)
    sys.stdout.write(f"{schema.as_str()}\n")



import strawberry


def a_resolver() -> list["AObject"]:
    return []


@strawberry.type
class ABase:
    a_name: str


@strawberry.type
class AObject(ABase):
    a_age: int

    @strawberry.field
    def a_is_of_full_age(self) -> bool:
        return self.a_age >= 18



from inline_snapshot import snapshot

from .utils.marks import requires_mypy, requires_pyright, requires_ty, skip_on_windows
from .utils.typecheck import Result, typecheck

pytestmark = [skip_on_windows, requires_pyright, requires_mypy, requires_ty]

CODE_ROUTER_WITH_CONTEXT = """
import strawberry

from strawberry.fastapi import GraphQLRouter, BaseContext


@strawberry.type
class Query:
    @strawberry.field
    def hello(self) -> str:
        return "Hello World"


class Context(BaseContext):
    pass


def get_context() -> Context:
    return Context()


router = GraphQLRouter(
    strawberry.Schema(
        query=Query,
    ),
    context_getter=get_context,
)

reveal_type(router)
"""


def test_router_with_context():
    results = typecheck(CODE_ROUTER_WITH_CONTEXT)

    assert results.pyright == snapshot(
        [
            Result(
                type="information",
                message='Type of "router" is "GraphQLRouter[Context, None]"',
                line=29,
                column=13,
            ),
        ]
    )
    assert results.mypy == snapshot(
        [
            Result(
                type="note",
                message='Revealed type is "strawberry.fastapi.router.GraphQLRouter[mypy_test.Context, None]"',
                line=29,
                column=13,
            ),
        ]
    )
    assert results.ty == snapshot(
        [
            Result(
                type="information",
                message="Revealed type: `GraphQLRouter[Context, None]`",
                line=29,
                column=13,
            ),
        ]
    )


CODE_ROUTER_WITH_ASYNC_CONTEXT = """
import strawberry

from strawberry.fastapi import GraphQLRouter, BaseContext


@strawberry.type
class Query:
    @strawberry.field
    def hello(self) -> str:
        return "Hello World"


class Context(BaseContext):
    pass


async def get_context() -> Context:
    return Context()


router = GraphQLRouter[Context](
    strawberry.Schema(
        query=Query,
    ),
    context_getter=get_context,
)

reveal_type(router)
"""


def test_router_with_async_context():
    results = typecheck(CODE_ROUTER_WITH_ASYNC_CONTEXT)

    assert results.pyright == snapshot(
        [
            Result(
                type="information",
                message='Type of "router" is "GraphQLRouter[Context, None]"',
                line=29,
                column=13,
            ),
        ]
    )
    assert results.mypy == snapshot(
        [
            Result(
                type="note",
                message='Revealed type is "strawberry.fastapi.router.GraphQLRouter[mypy_test.Context, None]"',
                line=29,
                column=13,
            ),
        ]
    )
    assert results.ty == snapshot(
        [
            Result(
                type="information",
                message="Revealed type: `GraphQLRouter[Context, None]`",
                line=29,
                column=13,
            ),
        ]
    )



from __future__ import annotations

from typing import TYPE_CHECKING, Annotated

import strawberry

if TYPE_CHECKING:
    from tests.b import B


@strawberry.type
class A:
    id: strawberry.ID

    @strawberry.field
    async def b(self) -> Annotated[B, strawberry.lazy("tests.b")]:
        from tests.b import B

        return B(id=self.id)

    @strawberry.field
    async def optional_b(self) -> Annotated[B, strawberry.lazy("tests.b")] | None:
        from tests.b import B

        return B(id=self.id)

    @strawberry.field
    async def optional_b2(self) -> Annotated[B, strawberry.lazy("tests.b")] | None:
        from tests.b import B

        return B(id=self.id)



from typing import Any
from unittest.mock import MagicMock

import pytest
from opentelemetry.trace import SpanKind
from pytest_mock import MockerFixture

import strawberry
from strawberry.extensions.tracing.opentelemetry import (
    OpenTelemetryExtension,
    OpenTelemetryExtensionSync,
)


@pytest.fixture
def global_tracer_mock(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("strawberry.extensions.tracing.opentelemetry.trace.get_tracer")


@strawberry.type
class Person:
    name: str = "Jess"


@strawberry.type
class Query:
    @strawberry.field
    async def person(self) -> Person:
        return Person()


@pytest.mark.asyncio
async def test_opentelemetry_uses_global_tracer(global_tracer_mock):
    schema = strawberry.Schema(query=Query, extensions=[OpenTelemetryExtension])

    query = """
        query {
            person {
                name
            }
        }
    """

    await schema.execute(query)

    global_tracer_mock.assert_called_once_with("strawberry", tracer_provider=None)


@pytest.mark.asyncio
async def test_opentelemetry_sync_uses_global_tracer(global_tracer_mock):
    schema = strawberry.Schema(query=Query, extensions=[OpenTelemetryExtensionSync])

    query = """
        query {
            person {
                name
            }
        }
    """

    await schema.execute(query)

    global_tracer_mock.assert_called_once_with("strawberry", tracer_provider=None)


def _instrumentation_stages(mocker, query):
    return [
        mocker.call("GraphQL Query", kind=SpanKind.SERVER),
        mocker.call().set_attribute("component", "graphql"),
        mocker.call().set_attribute("query", query),
        mocker.call("GraphQL Parsing", context=mocker.ANY),
        mocker.call().end(),
        mocker.call("GraphQL Validation", context=mocker.ANY),
        mocker.call().end(),
        mocker.call().end(),
    ]


@pytest.mark.asyncio
async def test_open_tracing(global_tracer_mock, mocker):
    schema = strawberry.Schema(query=Query, extensions=[OpenTelemetryExtension])
    query = """
        query {
            person {
                name
            }
        }
    """

    await schema.execute(query)

    # start_span is called by the Extension framework to instrument
    # phases of pre-request handling logic; parsing, validation, etc
    global_tracer_mock.return_value.start_span.assert_has_calls(
        _instrumentation_stages(mocker, query)
    )

    # start_as_current_span is called at the very start of request handling
    # it is a context manager, all other spans are a child of this
    global_tracer_mock.return_value.start_as_current_span.assert_has_calls(
        [
            mocker.call("GraphQL Resolving: person", context=mocker.ANY),
            mocker.call().__enter__(),
            mocker.call().__enter__().set_attribute("component", "graphql"),
            mocker.call().__enter__().set_attribute("graphql.parentType", "Query"),
            mocker.call().__enter__().set_attribute("graphql.path", "person"),
            mocker.call().__exit__(None, None, None),
        ]
    )


@pytest.mark.asyncio
async def test_open_tracing_uses_operation_name(global_tracer_mock, mocker):
    schema = strawberry.Schema(query=Query, extensions=[OpenTelemetryExtension])
    query = """
        query Example {
            person {
                name
            }
        }
    """

    await schema.execute(query, operation_name="Example")

    global_tracer_mock.return_value.start_span.assert_has_calls(
        [
            # if operation_name is supplied it is added to this span's tag
            mocker.call("GraphQL Query: Example", kind=SpanKind.SERVER),
            *_instrumentation_stages(mocker, query)[1:],
        ]
    )


@pytest.mark.asyncio
async def test_open_tracing_gets_operation_name(global_tracer_mock, mocker):
    schema = strawberry.Schema(query=Query, extensions=[OpenTelemetryExtension])
    query = """
        query Example {
            person {
                name
            }
        }
    """

    tracers = []

    def generate_trace(*args: str, **kwargs: Any):
        nonlocal tracers
        tracer = mocker.Mock()
        tracers.append(tracer)
        return tracer

    global_tracer_mock.return_value.start_span.side_effect = generate_trace

    await schema.execute(query)

    tracers[0].update_name.assert_has_calls(
        [
            # if operation_name is supplied it is added to this span's tag
            mocker.call("GraphQL Query: Example"),
        ]
    )


@pytest.mark.asyncio
async def test_tracing_add_kwargs(global_tracer_mock, mocker):
    @strawberry.type
    class Query:
        @strawberry.field
        def hi(self, name: str) -> str:
            return f"Hi {name}"

    schema = strawberry.Schema(query=Query, extensions=[OpenTelemetryExtension])

    query = """
        query {
            hi(name: "Patrick")
        }
    """

    await schema.execute(query)

    global_tracer_mock.return_value.start_as_current_span.assert_has_calls(
        [
            mocker.call().__enter__().set_attribute("graphql.parentType", "Query"),
            mocker.call().__enter__().set_attribute("graphql.path", "hi"),
            mocker.call().__enter__().set_attribute("graphql.param.name", "Patrick"),
        ]
    )


@pytest.mark.asyncio
async def test_tracing_filter_kwargs(global_tracer_mock, mocker):
    def arg_filter(kwargs, info):
        return {"name": "[...]"}

    @strawberry.type
    class Query:
        @strawberry.field
        def hi(self, name: str) -> str:
            return f"Hi {name}"

    schema = strawberry.Schema(
        query=Query, extensions=[OpenTelemetryExtension(arg_filter=arg_filter)]
    )

    query = """
        query {
            hi(name: "Patrick")
        }
    """

    await schema.execute(query)

    global_tracer_mock.return_value.start_as_current_span.assert_has_calls(
        [
            mocker.call().__enter__().set_attribute("graphql.parentType", "Query"),
            mocker.call().__enter__().set_attribute("graphql.path", "hi"),
            mocker.call().__enter__().set_attribute("graphql.param.name", "[...]"),
        ]
    )



import strawberry


# TODO: move this to common tests
def test_set_custom_http_response_status():
    from litestar import Litestar
    from litestar.testing import TestClient

    from strawberry.litestar import make_graphql_controller

    @strawberry.type
    class Query:
        @strawberry.field
        def abc(self, info: strawberry.Info) -> str:
            assert info.context.get("response") is not None
            info.context["response"].status_code = 418
            return "abc"

    schema = strawberry.Schema(query=Query)
    graphql_controller = make_graphql_controller(path="/graphql", schema=schema)
    app = Litestar(route_handlers=[graphql_controller])

    test_client = TestClient(app)
    response = test_client.post("/graphql", json={"query": "{ abc }"})
    assert response.status_code == 418
    assert response.json() == {"data": {"abc": "abc"}}


def test_set_without_setting_http_response_status():
    from litestar import Litestar
    from litestar.testing import TestClient

    from strawberry.litestar import make_graphql_controller

    @strawberry.type
    class Query:
        @strawberry.field
        def abc(self) -> str:
            return "abc"

    schema = strawberry.Schema(query=Query)
    graphql_controller = make_graphql_controller(path="/graphql", schema=schema)
    app = Litestar(route_handlers=[graphql_controller])

    test_client = TestClient(app)
    response = test_client.post("/graphql", json={"query": "{ abc }"})

    assert response.status_code == 200
    assert response.json() == {"data": {"abc": "abc"}}



import base64
from unittest.mock import Mock, patch

import pytest

import strawberry
from strawberry.extensions.tracing.apollo_federation import (
    ApolloFederationTracingExtension,
    ApolloFederationTracingExtensionSync,
)


@pytest.fixture
def mock_request_with_ftv1_header():
    """Create a mock request with FTV1 header."""
    request = Mock()
    request.headers = {"apollo-federation-include-trace": "ftv1"}
    return request


@pytest.fixture
def mock_request_without_header():
    """Create a mock request without FTV1 header."""
    request = Mock()
    request.headers = {}
    return request


def test_tracing_sync_with_ftv1_header(mock_request_with_ftv1_header):
    """Test that tracing activates with FTV1 header."""

    @strawberry.type
    class Person:
        name: str = "Alice"

    @strawberry.type
    class Query:
        @strawberry.field
        def person(self) -> Person:
            return Person()

    schema = strawberry.Schema(
        query=Query, extensions=[ApolloFederationTracingExtensionSync]
    )

    query = """
        query {
            person {
                name
            }
        }
    """

    result = schema.execute_sync(
        query, context_value={"request": mock_request_with_ftv1_header}
    )

    assert not result.errors
    assert "ftv1" in result.extensions
    # Verify it's base64 encoded
    ftv1_data = result.extensions["ftv1"]
    assert isinstance(ftv1_data, str)
    # Should be able to decode base64
    decoded = base64.b64decode(ftv1_data)
    assert isinstance(decoded, bytes)
    assert len(decoded) > 0


def test_tracing_sync_without_header(mock_request_without_header):
    """Test that tracing does not activate without FTV1 header."""

    @strawberry.type
    class Person:
        name: str = "Bob"

    @strawberry.type
    class Query:
        @strawberry.field
        def person(self) -> Person:
            return Person()

    schema = strawberry.Schema(
        query=Query, extensions=[ApolloFederationTracingExtensionSync]
    )

    query = """
        query {
            person {
                name
            }
        }
    """

    result = schema.execute_sync(
        query, context_value={"request": mock_request_without_header}
    )

    assert not result.errors
    # Should not have ftv1 in extensions
    assert "ftv1" not in result.extensions


@pytest.mark.asyncio
async def test_tracing_async_with_ftv1_header(mock_request_with_ftv1_header):
    """Test async tracing with FTV1 header."""

    @strawberry.type
    class Person:
        name: str = "Charlie"

    @strawberry.type
    class Query:
        @strawberry.field
        async def person(self) -> Person:
            return Person()

    schema = strawberry.Schema(
        query=Query, extensions=[ApolloFederationTracingExtension]
    )

    query = """
        query {
            person {
                name
            }
        }
    """

    result = await schema.execute(
        query, context_value={"request": mock_request_with_ftv1_header}
    )

    assert not result.errors
    assert "ftv1" in result.extensions
    ftv1_data = result.extensions["ftv1"]
    assert isinstance(ftv1_data, str)
    decoded = base64.b64decode(ftv1_data)
    assert isinstance(decoded, bytes)
    assert len(decoded) > 0


@pytest.mark.asyncio
async def test_tracing_async_without_header(mock_request_without_header):
    """Test async tracing without FTV1 header."""

    @strawberry.type
    class Person:
        name: str = "Diana"

    @strawberry.type
    class Query:
        @strawberry.field
        async def person(self) -> Person:
            return Person()

    schema = strawberry.Schema(
        query=Query, extensions=[ApolloFederationTracingExtension]
    )

    query = """
        query {
            person {
                name
            }
        }
    """

    result = await schema.execute(
        query, context_value={"request": mock_request_without_header}
    )

    assert not result.errors
    assert "ftv1" not in result.extensions


def test_tracing_with_nested_fields(mock_request_with_ftv1_header):
    """Test tracing with nested field resolvers."""

    @strawberry.type
    class Address:
        street: str = "Main St"
        city: str = "Springfield"

    @strawberry.type
    class Person:
        name: str = "Eve"

        @strawberry.field
        def address(self) -> Address:
            return Address()

    @strawberry.type
    class Query:
        @strawberry.field
        def person(self) -> Person:
            return Person()

    schema = strawberry.Schema(
        query=Query, extensions=[ApolloFederationTracingExtensionSync]
    )

    query = """
        query {
            person {
                name
                address {
                    street
                    city
                }
            }
        }
    """

    result = schema.execute_sync(
        query, context_value={"request": mock_request_with_ftv1_header}
    )

    assert not result.errors
    assert "ftv1" in result.extensions


def test_tracing_with_context_dict(mock_request_with_ftv1_header):
    """Test header detection with dict-style context."""

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "world"

    schema = strawberry.Schema(
        query=Query, extensions=[ApolloFederationTracingExtensionSync]
    )

    result = schema.execute_sync(
        "query { hello }", context_value={"request": mock_request_with_ftv1_header}
    )

    assert not result.errors
    assert "ftv1" in result.extensions


def test_tracing_with_context_object():
    """Test header detection with object-style context."""

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "world"

    schema = strawberry.Schema(
        query=Query, extensions=[ApolloFederationTracingExtensionSync]
    )

    # Create a context object with request attribute
    context = type("Context", (), {})()
    request = Mock()
    request.headers = {"apollo-federation-include-trace": "ftv1"}
    context.request = request

    result = schema.execute_sync("query { hello }", context_value=context)

    assert not result.errors
    assert "ftv1" in result.extensions


def test_tracing_handles_missing_protobuf():
    """Test graceful handling when protobuf is not installed."""

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "world"

    schema = strawberry.Schema(
        query=Query, extensions=[ApolloFederationTracingExtensionSync]
    )

    request = Mock()
    request.headers = {"apollo-federation-include-trace": "ftv1"}

    with patch(
        "strawberry.extensions.tracing.apollo_federation._check_protobuf_available",
        side_effect=ImportError("protobuf not installed"),
    ):
        result = schema.execute_sync(
            "query { hello }", context_value={"request": request}
        )

        # Should not error, just not include ftv1
        assert not result.errors
        assert "ftv1" not in result.extensions


def test_tracing_with_case_insensitive_header():
    """Test that header matching is case-insensitive."""

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "world"

    schema = strawberry.Schema(
        query=Query, extensions=[ApolloFederationTracingExtensionSync]
    )

    request = Mock()
    # Mix of uppercase and lowercase
    request.headers = {"Apollo-Federation-Include-Trace": "FTV1"}

    result = schema.execute_sync("query { hello }", context_value={"request": request})

    assert not result.errors
    assert "ftv1" in result.extensions


def test_tracing_with_wrong_header_value():
    """Test that tracing only activates with correct header value."""

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "world"

    schema = strawberry.Schema(
        query=Query, extensions=[ApolloFederationTracingExtensionSync]
    )

    request = Mock()
    request.headers = {"apollo-federation-include-trace": "wrong-value"}

    result = schema.execute_sync("query { hello }", context_value={"request": request})

    assert not result.errors
    assert "ftv1" not in result.extensions


def test_tracing_with_no_context():
    """Test that extension handles missing context gracefully."""

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "world"

    schema = strawberry.Schema(
        query=Query, extensions=[ApolloFederationTracingExtensionSync]
    )

    result = schema.execute_sync("query { hello }", context_value=None)

    assert not result.errors
    assert "ftv1" not in result.extensions


def test_tracing_with_no_request_in_context():
    """Test that extension handles missing request in context."""

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "world"

    schema = strawberry.Schema(
        query=Query, extensions=[ApolloFederationTracingExtensionSync]
    )

    result = schema.execute_sync("query { hello }", context_value={"other": "data"})

    assert not result.errors
    assert "ftv1" not in result.extensions


def test_lazy_import_from_public_tracing_module(mock_request_with_ftv1_header):
    """The extensions must be importable via the public lazy `tracing` module."""
    from strawberry.extensions import tracing

    assert tracing.ApolloFederationTracingExtension is ApolloFederationTracingExtension
    assert (
        tracing.ApolloFederationTracingExtensionSync
        is ApolloFederationTracingExtensionSync
    )

    @strawberry.type
    class Query:
        @strawberry.field
        def hello(self) -> str:
            return "world"

    schema = strawberry.Schema(
        query=Query, extensions=[tracing.ApolloFederationTracingExtensionSync]
    )

    result = schema.execute_sync(
        "query { hello }",
        context_value={"request": mock_request_with_ftv1_header},
    )

    assert not result.errors
    assert "ftv1" in result.extensions


def test_tracing_captures_errors_sync(mock_request_with_ftv1_header):
    """Test that errors are captured in FTV1 trace (sync)."""

    @strawberry.type
    class Query:
        @strawberry.field
        def failing(self) -> str:
            raise ValueError("Something went wrong")

    schema = strawberry.Schema(
        query=Query, extensions=[ApolloFederationTracingExtensionSync]
    )

    result = schema.execute_sync(
        "query { failing }", context_value={"request": mock_request_with_ftv1_header}
    )

    # Should have errors in the result
    assert result.errors is not None
    assert len(result.errors) == 1
    assert "Something went wrong" in str(result.errors[0])

    # Should still have ftv1 trace with error encoded
    assert "ftv1" in result.extensions
    ftv1_data = result.extensions["ftv1"]
    decoded = base64.b64decode(ftv1_data)
    assert isinstance(decoded, bytes)
    assert len(decoded) > 0
    # The error message should be in the trace
    assert b"Something went wrong" in decoded


@pytest.mark.asyncio
async def test_tracing_captures_errors_async(mock_request_with_ftv1_header):
    """Test that errors are captured in FTV1 trace (async)."""

    @strawberry.type
    class Query:
        @strawberry.field
        async def failing(self) -> str:
            raise ValueError("Async error occurred")

    schema = strawberry.Schema(
        query=Query, extensions=[ApolloFederationTracingExtension]
    )

    result = await schema.execute(
        "query { failing }", context_value={"request": mock_request_with_ftv1_header}
    )

    # Should have errors in the result
    assert result.errors is not None
    assert len(result.errors) == 1
    assert "Async error occurred" in str(result.errors[0])

    # Should still have ftv1 trace with error encoded
    assert "ftv1" in result.extensions
    ftv1_data = result.extensions["ftv1"]
    decoded = base64.b64decode(ftv1_data)
    assert isinstance(decoded, bytes)
    assert len(decoded) > 0
    # The error message should be in the trace
    assert b"Async error occurred" in decoded


def test_tracing_captures_nested_errors(mock_request_with_ftv1_header):
    """Test that errors in nested resolvers are captured."""

    @strawberry.type
    class Person:
        name: str = "Alice"

        @strawberry.field
        def failing_field(self) -> str:
            raise RuntimeError("Nested resolver failed")

    @strawberry.type
    class Query:
        @strawberry.field
        def person(self) -> Person:
            return Person()

    schema = strawberry.Schema(
        query=Query, extensions=[ApolloFederationTracingExtensionSync]
    )

    query = """
        query {
            person {
                name
                failingField
            }
        }
    """

    result = schema.execute_sync(
        query, context_value={"request": mock_request_with_ftv1_header}
    )

    assert result.errors is not None
    assert len(result.errors) == 1
    assert "Nested resolver failed" in str(result.errors[0])

    # Should still have ftv1 trace
    assert "ftv1" in result.extensions
    decoded = base64.b64decode(result.extensions["ftv1"])
    assert b"Nested resolver failed" in decoded


def test_tracing_error_json_includes_path(mock_request_with_ftv1_header):
    """Test that error JSON includes the field path."""

    @strawberry.type
    class Person:
        name: str = "Alice"

        @strawberry.field
        def failing_field(self) -> str:
            raise RuntimeError("Path test error")

    @strawberry.type
    class Query:
        @strawberry.field
        def person(self) -> Person:
            return Person()

    schema = strawberry.Schema(
        query=Query, extensions=[ApolloFederationTracingExtensionSync]
    )

    query = """
        query {
            person {
                failingField
            }
        }
    """

    result = schema.execute_sync(
        query, context_value={"request": mock_request_with_ftv1_header}
    )

    assert result.errors is not None
    decoded = base64.b64decode(result.extensions["ftv1"])

    # Verify JSON field contains path information
    assert b'"path"' in decoded
    assert b"failingField" in decoded
    assert b"person" in decoded


def test_simple_error_serialization():
    """Test _SimpleError serializes all fields correctly."""
    from strawberry.extensions.tracing.apollo_federation import _SimpleError

    error = _SimpleError(
        message="Test error",
        location_line=5,
        location_column=10,
        json_error='{"message":"Test error","path":["query","field"]}',
    )
    serialized = error.SerializeToString()

    # Verify message is in serialized output
    assert b"Test error" in serialized
    # Verify JSON content is in serialized output
    assert b'"path"' in serialized
    assert b'"message"' in serialized


def test_tracing_captures_multiple_errors(mock_request_with_ftv1_header):
    """Test that multiple errors in a single query are all captured."""

    @strawberry.type
    class Query:
        @strawberry.field
        def failing_one(self) -> str | None:
            raise ValueError("First error")

        @strawberry.field
        def failing_two(self) -> str | None:
            raise ValueError("Second error")

    schema = strawberry.Schema(
        query=Query, extensions=[ApolloFederationTracingExtensionSync]
    )

    query = """
        query {
            failingOne
            failingTwo
        }
    """

    result = schema.execute_sync(
        query, context_value={"request": mock_request_with_ftv1_header}
    )

    # Should have both errors (nullable fields allow independent failures)
    assert result.errors is not None
    assert len(result.errors) == 2

    # Both error messages should be in the trace
    assert "ftv1" in result.extensions
    decoded = base64.b64decode(result.extensions["ftv1"])
    assert b"First error" in decoded
    assert b"Second error" in decoded


def test_tracing_error_location_encoding(mock_request_with_ftv1_header):
    """Test that error location (line/column) is encoded in FTV1 trace."""
    from strawberry.extensions.tracing.apollo_federation import (
        _SimpleError,
    )

    # Unit test: verify location encoding in _SimpleError
    error = _SimpleError(
        message="Location test",
        location_line=3,
        location_column=15,
        json_error='{"message":"Location test"}',
    )
    serialized = error.SerializeToString()

    # Field 2 (location) should be present with wire type 2 (length-delimited)
    # 0x12 = (2 << 3) | 2 = field 2, wire type 2
    assert b"\x12" in serialized

    # Integration test: verify location is captured from actual query
    @strawberry.type
    class Query:
        @strawberry.field
        def failing(self) -> str:
            raise ValueError("Location error")

    schema = strawberry.Schema(
        query=Query, extensions=[ApolloFederationTracingExtensionSync]
    )

    result = schema.execute_sync(
        "query { failing }", context_value={"request": mock_request_with_ftv1_header}
    )

    assert result.errors is not None
    decoded = base64.b64decode(result.extensions["ftv1"])

    # The location field marker (0x12) should be in the trace
    assert b"\x12" in decoded
    # The JSON should contain locations array
    assert b'"locations"' in decoded
    assert b'"line"' in decoded
    assert b'"column"' in decoded



import asyncio

import pytest
from pytest_codspeed.plugin import BenchmarkFixture

from .schema import schema

query = """
fragment AuthorFragment on User {
  id
  username
  email
  role
}

fragment PostFragment on Post {
  id
  content
  title

  author {
    ...AuthorFragment
  }

  comments {
    edges {
      node {
        author {
          id
          username
          email
          role
        }
      }
    }
  }
}

query Query($query: String!, $first: Int!) {
  search(query: $query, first: $first) {
    ... on User {
      id
      username
      email
      role
      posts {
        pageInfo {
          hasNextPage
          hasPreviousPage
          startCursor
          endCursor
        }
        edges {
          cursor
          node {
            author {
              email
              id
              posts {
                edges {
                  node {
                    ...PostFragment
                  }
                }
              }
            }
          }
        }
      }
    }
    ... on Post {
      ...PostFragment
    }
    ... on Comment {
      id
      text
      author {
        ...AuthorFragment
      }
      post {
        ...PostFragment
      }
    }
  }
}
"""


@pytest.mark.parametrize("number", [50])
def test_execute_complex_schema(benchmark: BenchmarkFixture, number: int):
    def run():
        coroutine = schema.execute(
            query,
            variable_values={"query": "test", "first": number},
        )

        return asyncio.run(coroutine)

    result = benchmark(run)

    assert not result.errors



import pytest

import strawberry
from strawberry.federation.schema_directives import Context, Cost, ListSize, Policy


@pytest.mark.parametrize(
    ("directive_name", "minimum_version", "valid_version", "invalid_version"),
    [
        ("policy", "2.6", "2.6", "2.5"),
        ("context", "2.8", "2.8", "2.7"),
        ("cost", "2.9", "2.9", "2.8"),
        ("listSize", "2.9", "2.9", "2.8"),
    ],
)
def test_directive_version_validation(
    directive_name: str,
    minimum_version: str,
    valid_version: str,
    invalid_version: str,
):
    """Test that directives validate their minimum federation version requirement"""

    # Create type with directive based on directive name
    if directive_name == "policy":

        @strawberry.federation.type
        class ProductPolicy:
            upc: str
            name: str = strawberry.federation.field(
                directives=[Policy(policies=[["admin"]])]
            )

        @strawberry.type
        class QueryPolicy:
            product: ProductPolicy

        query_type = QueryPolicy

    elif directive_name == "context":

        @strawberry.federation.type(directives=[Context(name="userContext")])
        class UserContext:
            id: strawberry.ID
            name: str

        @strawberry.type
        class QueryContext:
            user: UserContext

        query_type = QueryContext

    elif directive_name == "cost":

        @strawberry.federation.type
        class ProductCost:
            upc: str
            name: str = strawberry.federation.field(directives=[Cost(weight=10)])

        @strawberry.type
        class QueryCost:
            product: ProductCost

        query_type = QueryCost

    else:  # listSize

        @strawberry.federation.type
        class ProductListSize:
            upc: str
            friends: list[str] = strawberry.federation.field(
                directives=[
                    ListSize(
                        assumed_size=100,
                        slicing_arguments=None,
                        sized_fields=None,
                    )
                ]
            )

        @strawberry.type
        class QueryListSize:
            product: ProductListSize

        query_type = QueryListSize

    # Should work with valid version
    schema = strawberry.federation.Schema(
        query=query_type,
        federation_version=valid_version,  # type: ignore
    )
    assert schema is not None

    # Should fail with invalid version
    escaped_version = minimum_version.replace(".", r"\.")
    expected_error = (
        f"Directive @{directive_name} requires federation version "
        f"v{escaped_version} or higher"
    )
    with pytest.raises(ValueError, match=expected_error):
        strawberry.federation.Schema(
            query=query_type,
            federation_version=invalid_version,  # type: ignore
        )


def test_directive_version_validation_multiple_directives():
    """Test validation with multiple directives having different version requirements"""

    @strawberry.federation.type(directives=[Context(name="ctx")])
    class Product:
        upc: str
        name: str = strawberry.federation.field(
            directives=[Policy(policies=[["admin"]]), Cost(weight=5)]
        )

    @strawberry.type
    class Query:
        product: Product

    # Should work with v2.9+ (highest requirement among @context v2.8, @policy v2.6, @cost v2.9)
    schema = strawberry.federation.Schema(query=Query, federation_version="2.9")
    assert schema is not None

    # Should fail with v2.8 (cost requires v2.9)
    with pytest.raises(
        ValueError, match=r"Directive @cost requires federation version v2\.9 or higher"
    ):
        strawberry.federation.Schema(query=Query, federation_version="2.8")

    # Should fail with v2.7 (context requires v2.8)
    with pytest.raises(
        ValueError,
        match=r"Directive @context requires federation version v2\.8 or higher",
    ):
        strawberry.federation.Schema(query=Query, federation_version="2.7")


def test_older_directives_work_with_any_version():
    """Test that older directives work with any federation version"""

    @strawberry.federation.type(keys=["id"])
    class Product:
        id: strawberry.ID
        name: str = strawberry.federation.field(external=True)

    @strawberry.type
    class Query:
        product: Product

    # Should work with any version since @key and @external are from v2.0
    for version in ["2.0", "2.5", "2.7", "2.9", "2.11"]:
        schema = strawberry.federation.Schema(query=Query, federation_version=version)  # type: ignore
        assert schema is not None


def test_default_version_uses_latest():
    """Test that default federation version uses latest"""

    @strawberry.federation.type
    class Product:
        upc: str

    @strawberry.type
    class Query:
        product: Product

    # Default should use latest version
    schema = strawberry.federation.Schema(query=Query)
    assert schema.federation_version == (2, 11)  # Latest version



import asyncio
from collections.abc import AsyncGenerator

import pytest
from freezegun import freeze_time
from graphql.utilities import get_introspection_query

import strawberry
from strawberry.extensions import SchemaExtension
from strawberry.extensions.tracing.apollo import (
    ApolloTracingExtension,
    ApolloTracingExtensionSync,
)


@freeze_time("20120114 12:00:01")
def test_tracing_sync(mocker):
    mocker.patch(
        "strawberry.extensions.tracing.apollo.time.perf_counter_ns", return_value=0
    )

    @strawberry.type
    class Person:
        name: str = "Jess"

    @strawberry.type
    class Query:
        @strawberry.field
        def person(self) -> Person:
            return Person()

    schema = strawberry.Schema(query=Query, extensions=[ApolloTracingExtensionSync])

    query = """
        query {
            person {
                name
            }
        }
    """

    result = schema.execute_sync(query)

    assert not result.errors

    assert result.extensions == {
        "tracing": {
            "version": 1,
            "startTime": "2012-01-14T12:00:01.000000Z",
            "endTime": "2012-01-14T12:00:01.000000Z",
            "duration": 0,
            "execution": {
                "resolvers": [
                    {
                        "path": ["person"],
                        "field_name": "person",
                        "parentType": "Query",
                        "returnType": "Person!",
                        "startOffset": 0,
                        "duration": 0,
                    },
                ]
            },
            "validation": {"startOffset": 0, "duration": 0},
            "parsing": {"startOffset": 0, "duration": 0},
        }
    }


@pytest.mark.asyncio
@freeze_time("20120114 12:00:01")
async def test_tracing_async(mocker):
    mocker.patch(
        "strawberry.extensions.tracing.apollo.time.perf_counter_ns", return_value=0
    )

    @strawberry.type
    class Person:
        name: str = "Jess"

    @strawberry.type
    class Query:
        @strawberry.field
        def example(self) -> str:
            return "Hi"

        @strawberry.field
        async def person(self) -> Person:
            return Person()

    schema = strawberry.Schema(query=Query, extensions=[ApolloTracingExtension])

    query = """
        query {
            example
            person {
                name
            }
        }
    """

    result = await schema.execute(query)

    assert not result.errors

    assert result.extensions == {
        "tracing": {
            "version": 1,
            "startTime": "2012-01-14T12:00:01.000000Z",
            "endTime": "2012-01-14T12:00:01.000000Z",
            "duration": 0,
            "execution": {
                "resolvers": [
                    {
                        "duration": 0,
                        "field_name": "example",
                        "parentType": "Query",
                        "path": ["example"],
                        "returnType": "String!",
                        "startOffset": 0,
                    },
                    {
                        "path": ["person"],
                        "field_name": "person",
                        "parentType": "Query",
                        "returnType": "Person!",
                        "startOffset": 0,
                        "duration": 0,
                    },
                ]
            },
            "validation": {"startOffset": 0, "duration": 0},
            "parsing": {"startOffset": 0, "duration": 0},
        }
    }


@freeze_time("20120114 12:00:01")
def test_should_not_trace_introspection_sync_queries(mocker):
    mocker.patch(
        "strawberry.extensions.tracing.apollo.time.perf_counter_ns", return_value=0
    )

    @strawberry.type
    class Person:
        name: str = "Jess"

    @strawberry.type
    class Query:
        @strawberry.field
        def person(self) -> Person:
            return Person()

    schema = strawberry.Schema(query=Query, extensions=[ApolloTracingExtensionSync])

    result = schema.execute_sync(get_introspection_query())

    assert not result.errors
    assert result.extensions == {
        "tracing": {
            "version": 1,
            "startTime": "2012-01-14T12:00:01.000000Z",
            "endTime": "2012-01-14T12:00:01.000000Z",
            "duration": 0,
            "execution": {"resolvers": []},
            "validation": {"startOffset": 0, "duration": 0},
            "parsing": {"startOffset": 0, "duration": 0},
        }
    }


@pytest.mark.asyncio
@freeze_time("20120114 12:00:01")
async def test_should_not_trace_introspection_async_queries(mocker):
    mocker.patch(
        "strawberry.extensions.tracing.apollo.time.perf_counter_ns", return_value=0
    )

    @strawberry.type
    class Person:
        name: str = "Jess"

    @strawberry.type
    class Query:
        @strawberry.field
        async def person(self) -> Person:
            return Person()

    schema = strawberry.Schema(query=Query, extensions=[ApolloTracingExtension])

    result = await schema.execute(get_introspection_query())

    assert not result.errors
    assert result.extensions == {
        "tracing": {
            "version": 1,
            "startTime": "2012-01-14T12:00:01.000000Z",
            "endTime": "2012-01-14T12:00:01.000000Z",
            "duration": 0,
            "execution": {"resolvers": []},
            "validation": {"startOffset": 0, "duration": 0},
            "parsing": {"startOffset": 0, "duration": 0},
        }
    }


@pytest.mark.asyncio
async def test_tracing_resolvers_populated_on_multiple_executions():
    """Test that resolvers field is populated on every execution, not just the first."""

    @strawberry.type
    class Query:
        @strawberry.field
        def node(self) -> str:
            return ""

    schema = strawberry.Schema(Query, extensions=[ApolloTracingExtension])

    for i in range(3):
        result = await schema.execute("{ node }")
        assert not result.errors
        assert result.extensions is not None
        resolvers = result.extensions["tracing"]["execution"]["resolvers"]
        assert len(resolvers) == 1, (
            f"Expected 1 resolver on execution {i}, got {len(resolvers)}"
        )
        assert resolvers[0]["field_name"] == "node"


def test_tracing_resolvers_populated_on_multiple_sync_executions():
    """Test that resolvers field is populated on every sync execution, not just the first."""

    @strawberry.type
    class Query:
        @strawberry.field
        def node(self) -> str:
            return ""

    schema = strawberry.Schema(Query, extensions=[ApolloTracingExtensionSync])

    for i in range(3):
        result = schema.execute_sync("{ node }")
        assert not result.errors
        assert result.extensions is not None
        resolvers = result.extensions["tracing"]["execution"]["resolvers"]
        assert len(resolvers) == 1, (
            f"Expected 1 resolver on execution {i}, got {len(resolvers)}"
        )
        assert resolvers[0]["field_name"] == "node"


def test_tracing_invalid_query_sync():
    """Regression test: invalid query must not crash ApolloTracingExtension."""

    @strawberry.type
    class Query:
        @strawberry.field
        def node(self) -> str:
            return ""

    schema = strawberry.Schema(Query, extensions=[ApolloTracingExtensionSync])

    result = schema.execute_sync("{ node() }")
    assert result.errors
    assert result.extensions is not None
    assert "tracing" in result.extensions


@pytest.mark.asyncio
async def test_tracing_invalid_query_async():
    """Regression test: invalid query must not crash ApolloTracingExtension."""

    @strawberry.type
    class Query:
        @strawberry.field
        def node(self) -> str:
            return ""

    schema = strawberry.Schema(Query, extensions=[ApolloTracingExtension])

    result = await schema.execute("{ node() }")
    assert result.errors
    assert result.extensions is not None
    assert "tracing" in result.extensions


@pytest.mark.asyncio
async def test_concurrent_execution_context_result():
    """Regression test: execution_context.result must not be None during concurrent requests."""
    results = []

    class RecordResultExtension(SchemaExtension):
        async def on_operation(self):
            yield
            results.append(self.execution_context.result)

    @strawberry.type
    class Query:
        @strawberry.field
        def node(self) -> str:
            return ""

    schema = strawberry.Schema(Query, extensions=[RecordResultExtension])
    await asyncio.gather(
        schema.execute("query { node }"),
        schema.execute("query { node }"),
    )
    assert len(results) == 2
    assert all(r is not None for r in results)


@pytest.mark.asyncio
async def test_concurrent_subscribe_has_isolated_extensions():
    """Regression test: concurrent subscriptions must get isolated extension instances."""
    contexts: list[object] = []

    class RecordContextExtension(SchemaExtension):
        async def on_execute(self):
            contexts.append(id(self.execution_context))
            yield

    @strawberry.type
    class Query:
        @strawberry.field
        def node(self) -> str:
            return ""

    @strawberry.type
    class Subscription:
        @strawberry.subscription
        async def node(self) -> AsyncGenerator[str, None]:
            yield ""

    schema = strawberry.Schema(
        query=Query, subscription=Subscription, extensions=[RecordContextExtension]
    )

    async def run() -> None:
        sub = await schema.subscribe("subscription { node }")
        async for result in sub:  # type: ignore[union-attr]
            assert result.data == {"node": ""}
            break

    await asyncio.gather(run(), run())
    assert len(contexts) == 2
    assert contexts[0] != contexts[1], "each subscription must have its own context"



from __future__ import annotations

from io import BytesIO
from json import dumps
from typing import Any, Literal

from django.core.exceptions import BadRequest, SuspiciousOperation
from django.core.files.uploadedfile import SimpleUploadedFile
from django.http import Http404, HttpRequest, HttpResponse
from django.test.client import RequestFactory

from strawberry.django.views import GraphQLView as BaseGraphQLView
from strawberry.http import GraphQLHTTPResponse
from strawberry.http.ides import GraphQL_IDE
from strawberry.schema import Schema
from strawberry.types import ExecutionResult
from tests.http.context import get_context
from tests.views.schema import Query

from .base import JSON, HttpClient, Response, ResultOverrideFunction


class GraphQLView(BaseGraphQLView[dict[str, object], object]):
    result_override: ResultOverrideFunction = None

    def get_root_value(self, request) -> Query:
        super().get_root_value(request)  # for coverage
        return Query()

    def get_context(
        self, request: HttpRequest, response: HttpResponse
    ) -> dict[str, object]:
        context = {"request": request, "response": response}

        return get_context(context)

    def process_result(
        self, request: HttpRequest, result: ExecutionResult
    ) -> GraphQLHTTPResponse:
        if self.result_override:
            return self.result_override(result)

        return super().process_result(request, result)


class DjangoHttpClient(HttpClient):
    def __init__(
        self,
        schema: Schema,
        graphql_ide: GraphQL_IDE | None = "graphiql",
        allow_queries_via_get: bool = True,
        result_override: ResultOverrideFunction = None,
        multipart_uploads_enabled: bool = False,
    ):
        self.view = GraphQLView.as_view(
            schema=schema,
            graphql_ide=graphql_ide,
            allow_queries_via_get=allow_queries_via_get,
            result_override=result_override,
            multipart_uploads_enabled=multipart_uploads_enabled,
        )

    def _get_header_name(self, key: str) -> str:
        return f"HTTP_{key.upper().replace('-', '_')}"

    def _to_django_headers(self, headers: dict[str, str]) -> dict[str, str]:
        return {self._get_header_name(key): value for key, value in headers.items()}

    def _get_headers(
        self,
        method: Literal["get", "post"],
        headers: dict[str, str] | None,
        files: dict[str, BytesIO] | None,
    ) -> dict[str, str]:
        headers = headers or {}
        headers = self._to_django_headers(headers)

        return super()._get_headers(method=method, headers=headers, files=files)

    async def _do_request(self, request: HttpRequest) -> Response:
        try:
            response = self.view(request)
        except Http404:
            return Response(status_code=404, data=b"Not found")
        except (BadRequest, SuspiciousOperation) as e:
            return Response(status_code=400, data=e.args[0].encode())
        else:
            return Response(
                status_code=response.status_code,
                data=response.content,
                headers=dict(response.headers),
            )

    async def _graphql_request(
        self,
        method: Literal["get", "post"],
        query: str | None = None,
        operation_name: str | None = None,
        variables: dict[str, object] | None = None,
        files: dict[str, BytesIO] | None = None,
        headers: dict[str, str] | None = None,
        extensions: dict[str, Any] | None = None,
        **kwargs: Any,
    ) -> Response:
        headers = self._get_headers(method=method, headers=headers, files=files)
        additional_arguments = {**kwargs, **headers}

        body = self._build_body(
            query=query,
            operation_name=operation_name,
            variables=variables,
            files=files,
            method=method,
            extensions=extensions,
        )

        data: dict[str, object] | str | None = None

        if body and files:
            body.update(
                {
                    name: SimpleUploadedFile(name, file.read())
                    for name, file in files.items()
                }
            )
        else:
            additional_arguments["content_type"] = "application/json"

        if body:
            data = body if files or method == "get" else dumps(body)

        factory = RequestFactory()
        request = getattr(factory, method)(
            "/graphql",
            data=data,
            **additional_arguments,
        )

        return await self._do_request(request)

    async def request(
        self,
        url: str,
        method: Literal["head", "get", "post", "patch", "put", "delete"],
        headers: dict[str, str] | None = None,
    ) -> Response:
        headers = headers or {}

        factory = RequestFactory()
        request = getattr(factory, method)(url, **headers)

        return await self._do_request(request)

    async def get(
        self,
        url: str,
        headers: dict[str, str] | None = None,
    ) -> Response:
        django_headers = self._to_django_headers(headers or {})
        return await self.request(url, "get", headers=django_headers)

    async def post(
        self,
        url: str,
        data: bytes | None = None,
        json: JSON | None = None,
        headers: dict[str, str] | None = None,
    ) -> Response:
        headers = headers or {}
        content_type = headers.pop("Content-Type", "")

        body = dumps(json) if json is not None else data

        factory = RequestFactory()
        request = factory.post(
            url,
            data=body,
            content_type=content_type,
            headers=headers,
        )

        return await self._do_request(request)



import shutil
import sys

import pytest


def pyright_exist() -> bool:
    return shutil.which("pyright") is not None


def mypy_exists() -> bool:
    return shutil.which("mypy") is not None


def ty_exists() -> bool:
    return shutil.which("ty") is not None


skip_on_windows = pytest.mark.skipif(
    sys.platform == "win32",
    reason="Do not run pyright on windows due to path issues",
)

requires_pyright = pytest.mark.skipif(
    not pyright_exist(),
    reason="These tests require pyright",
)

requires_mypy = pytest.mark.skipif(
    not mypy_exists(),
    reason="These tests require mypy",
)

requires_ty = pytest.mark.skipif(
    not ty_exists(),
    reason="These tests require ty",
)



import strawberry


def test_type_add_type_definition_with_fields():
    @strawberry.type
    class Query:
        name: str
        age: int

    definition = Query.__strawberry_definition__

    assert definition.name == "Query"
    assert len(definition.fields) == 2

    assert definition.fields[0].python_name == "name"
    assert definition.fields[0].type is str

    assert definition.fields[1].python_name == "age"
    assert definition.fields[1].type is int


def test_passing_nothing_to_fields():
    @strawberry.type
    class Query:
        name: str = strawberry.field()
        age: int = strawberry.field()

    definition = Query.__strawberry_definition__

    assert definition.name == "Query"
    assert len(definition.fields) == 2

    assert definition.fields[0].python_name == "name"
    assert definition.fields[0].type is str

    assert definition.fields[1].python_name == "age"
    assert definition.fields[1].type is int



from __future__ import annotations

from typing import TYPE_CHECKING, Annotated

import strawberry

if TYPE_CHECKING:
    from tests.a import A


@strawberry.type
class B:
    id: strawberry.ID

    @strawberry.field
    async def a(self) -> Annotated[A, strawberry.lazy("tests.a"), object()]:
        from tests.a import A

        return A(id=self.id)

    @strawberry.field
    async def a_list(
        self,
    ) -> list[Annotated[A, strawberry.lazy("tests.a")]]:  # pragma: no cover
        from tests.a import A

        return [A(id=self.id)]

    @strawberry.field
    async def optional_a(
        self,
    ) -> Annotated[A, strawberry.lazy("tests.a"), object()] | None:
        from tests.a import A

        return A(id=self.id)

    @strawberry.field
    async def optional_a2(
        self,
    ) -> Annotated[A, strawberry.lazy("tests.a"), object()] | None:
        from tests.a import A

        return A(id=self.id)
